# ---------------------------- #
# --- Setup the simulation --- #
# ---------------------------- #

import sim_class, sim_agents
from abmarl.managers import SimulationManager
from abmarl.external import MultiAgentWrapper
from ray.tune.registry import register_env
sim_config = {
    # Fill in simulation configuration
}

sim_creator = lambda sim_config: MultiAgentWrapper(SimulationManager(sim_class(sim_config)))
sim = sim_creator(sim_config)
sim_name = "SimulationName"
register_env(sim_name, sim_creator)

# -------------------------- #
# --- Setup the policies --- #
# -------------------------- #

from abmarl.pols import HeuristicPolicy


class CustomHeuristicPolicy(HeuristicPolicy):
    """A custom heuristic policy for you the design"""
    def compute_actions(self, obs_batch, *args, **kwargs):
        # return [some_action for _ in obs_batch], [], {}
        return [0 for _ in obs_batch], [], {}


agents = sim.agents
policies = {
    'policy_0_name': (None, agents[0].observation_space, agents[0].action_space, {}),
    'policy_1_name': (None, agents[1].observation_space, agents[1].action_space, {}),
    'policy_2_name': (None, agents[2].observation_space, agents[2].action_space, {})
}


def policy_mapping_fn(agent_id):
    pass # Map the agent id to the policy you want that agent to train.


# --------------------------- #
# --- Setup the algorithm --- #
# --------------------------- #

# Full list of supported algorithms here:
# https://docs.ray.io/en/releases-1.2.0/rllib-algorithms.html#rllib-algorithms
algo_name = 'PG'


# ------------------ #
# --- Parameters --- #
# ------------------ #

# List of common ray_tune parameters here:
# https://docs.ray.io/en/releases-1.2.0/rllib-training.html#common-parameters
params = {
    'experiment': {
        'title': '{}'.format('The-title-of-this-experiment'),
        'sim_creator': sim_creator,
    },
    'ray_tune': {
        'run_or_experiment': algo_name,
        'stop': {
            # Stopping criteria
        },
        'config': {
            # --- Simulation ---
            'sim': sim_name,
            'sim_config': sim_config,
            # --- Multiagent ---
            'multiagent': {
                'policies': policies,
                'policy_mapping_fn': policy_mapping_fn,
            },
        },
    }
}



from gym.spaces import Box, MultiBinary, Discrete, Dict
import numpy as np

from abmarl.sim import Agent
from abmarl.sim.corridor import MultiCorridor as Corridor


def test_corridor_attributes():
    assert Corridor.Actions.LEFT == 0
    assert Corridor.Actions.RIGHT == 2
    assert Corridor.Actions.STAY == 1


def test_corridor_init():
    sim = Corridor()
    assert sim.end == 10
    assert sim.agents == {
        'agent0': Agent(
            id='agent0',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int),
                'left': MultiBinary(1),
                'right': MultiBinary(1)
            }), action_space=Discrete(3)),
        'agent1': Agent(
            id='agent1',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int),
                'left': MultiBinary(1),
                'right': MultiBinary(1)
            }), action_space=Discrete(3)),
        'agent2': Agent(
            id='agent2',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int),
                'left': MultiBinary(1),
                'right': MultiBinary(1)
            }), action_space=Discrete(3)),
        'agent3': Agent(
            id='agent3',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int),
                'left': MultiBinary(1),
                'right': MultiBinary(1)
            }), action_space=Discrete(3)),
        'agent4': Agent(
            id='agent4',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int),
                'left': MultiBinary(1),
                'right': MultiBinary(1)
            }), action_space=Discrete(3)),
    }


def test_corridor_init_end():
    sim = Corridor(end=7)
    assert sim.end == 7
    for agent in sim.agents.values():
        assert agent.observation_space['position'] == Box(0, 6, (1,), np.int)


def test_corridor_init_num_agents():
    sim = Corridor(num_agents=2)
    assert sim.agents == {
        'agent0': Agent(
            id='agent0',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int), 'left': MultiBinary(1), 'right': MultiBinary(1)
            }), action_space=Discrete(3)),
        'agent1': Agent(
            id='agent1',
            observation_space=Dict({
                'position': Box(0,9,(1,),np.int), 'left': MultiBinary(1), 'right': MultiBinary(1)
            }), action_space=Discrete(3)),
    }


def test_corridor_reset():
    np.random.seed(24)
    sim = Corridor()
    sim.reset()
    for agent in sim.agents.values():
        assert agent == sim.corridor[agent.position]
    assert sim.reward == {agent_id: 0 for agent_id in sim.agents}


def test_corridor_step():
    np.random.seed(24)
    sim = Corridor()
    sim.reset()
    assert sim.corridor[4].id == 'agent3'
    assert sim.corridor[5].id == 'agent4'
    assert sim.corridor[6].id == 'agent2'
    assert sim.corridor[7].id == 'agent1'
    assert sim.corridor[8].id == 'agent0'

    # Get observation and make some assertion
    assert sim.get_obs('agent0') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent1') == {'left': [True], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [6], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [True]}

    sim.step({'agent0': Corridor.Actions.RIGHT})
    assert sim.corridor[8] is None
    assert sim.corridor[9] is None
    assert sim.get_obs('agent0') == {'left': [False], 'position': [9], 'right': [False]}
    assert sim.get_obs('agent1') == {'left': [True], 'position': [7], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [6], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [True]}
    assert sim.get_reward('agent0') == 100
    assert sim.get_done('agent0')

    sim.step({'agent1': Corridor.Actions.RIGHT})
    assert sim.corridor[7] is None
    assert sim.corridor[8].id == 'agent1'
    assert sim.get_obs('agent1') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [6], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [True]}
    assert sim.get_reward('agent1') == -1
    assert not sim.get_done('agent1')

    sim.step({'agent2': Corridor.Actions.RIGHT})
    assert sim.corridor[6] is None
    assert sim.corridor[7].id == 'agent2'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [False], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent2') == -1
    assert not sim.get_done('agent2')

    sim.step({'agent3': Corridor.Actions.RIGHT})
    assert sim.corridor[4].id == 'agent3'
    assert sim.corridor[5].id == 'agent4'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [False], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent3') == -5
    assert not sim.get_done('agent3')

    sim.step({'agent4': Corridor.Actions.RIGHT})
    assert sim.corridor[5] is None
    assert sim.corridor[6].id == 'agent4'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [6], 'right': [True]}
    assert sim.get_reward('agent4') == -3
    assert not sim.get_done('agent4')

    sim.step({'agent1': Corridor.Actions.STAY})
    assert sim.corridor[8].id == 'agent1'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [6], 'right': [True]}
    assert sim.get_reward('agent1') == -1
    assert not sim.get_done('agent1')

    sim.step({'agent2': Corridor.Actions.LEFT})
    assert sim.corridor[7].id == 'agent2'
    assert sim.corridor[6].id == 'agent4'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [6], 'right': [True]}
    assert sim.get_reward('agent2') == -5
    assert not sim.get_done('agent2')

    sim.step({'agent3': Corridor.Actions.STAY})
    assert sim.corridor[4].id == 'agent3'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [True], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [6], 'right': [True]}
    assert sim.get_reward('agent3') == -1
    assert not sim.get_done('agent3')

    sim.step({'agent4': Corridor.Actions.LEFT})
    assert sim.corridor[5].id == 'agent4'
    assert sim.get_obs('agent1') == {'left': [True], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [False], 'position': [7], 'right': [True]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent4') == -3
    assert not sim.get_done('agent4')

    sim.step({'agent1': Corridor.Actions.RIGHT})
    assert sim.corridor[9] is None
    assert sim.get_obs('agent1') == {'left': [False], 'position': [9], 'right': [False]}
    assert sim.get_obs('agent2') == {'left': [False], 'position': [7], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent1') == 100
    assert sim.get_done('agent1')

    sim.step({'agent2': Corridor.Actions.RIGHT})
    assert sim.corridor[8].id == 'agent2'
    assert sim.corridor[7] is None
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent2') == -1
    assert not sim.get_done('agent2')

    sim.step({'agent3': Corridor.Actions.RIGHT})
    assert sim.corridor[4].id == 'agent3'
    assert sim.corridor[5].id == 'agent4'
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent3') == -5
    assert not sim.get_done('agent3')

    sim.step({'agent4': Corridor.Actions.LEFT})
    assert sim.corridor[4].id == 'agent3'
    assert sim.corridor[5].id == 'agent4'
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent4') == -7
    assert not sim.get_done('agent4')

    sim.step({'agent2': Corridor.Actions.STAY})
    assert sim.corridor[8].id == 'agent2'
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [4], 'right': [True]}
    assert sim.get_obs('agent4') == {'left': [True], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent2') == -1
    assert not sim.get_done('agent2')

    sim.step({'agent3': Corridor.Actions.LEFT})
    assert sim.corridor[4] is None
    assert sim.corridor[3].id == 'agent3'
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [3], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [5], 'right': [False]}
    assert sim.get_reward('agent3') == -3
    assert not sim.get_done('agent3')

    sim.step({'agent4': Corridor.Actions.RIGHT})
    assert sim.corridor[5] is None
    assert sim.corridor[6].id == 'agent4'
    assert sim.get_obs('agent2') == {'left': [False], 'position': [8], 'right': [False]}
    assert sim.get_obs('agent3') == {'left': [False], 'position': [3], 'right': [False]}
    assert sim.get_obs('agent4') == {'left': [False], 'position': [6], 'right': [False]}
    assert sim.get_reward('agent4') == -1
    assert not sim.get_done('agent4')

    assert not sim.get_all_done()
    sim.step({'agent2': Corridor.Actions.RIGHT})
    sim.step({'agent4': Corridor.Actions.RIGHT})
    sim.step({'agent4': Corridor.Actions.RIGHT})
    sim.step({'agent4': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    sim.step({'agent3': Corridor.Actions.RIGHT})
    assert sim.get_all_done()




from matplotlib import pyplot as plt
import numpy as np

from abmarl.sim.gridworld.base import GridWorldSimulation
from abmarl.sim.gridworld.agent import GridObservingAgent, MovingAgent, AttackingAgent, HealthAgent
from abmarl.sim.gridworld.state import HealthState, PositionState
from abmarl.sim.gridworld.actor import MoveActor, AttackActor
from abmarl.sim.gridworld.observer import SingleGridObserver
from abmarl.sim.gridworld.done import OneTeamRemainingDone


class BattleAgent(GridObservingAgent, MovingAgent, AttackingAgent, HealthAgent):
    def __init__(self, **kwargs):
        super().__init__(
            move_range=1,
            attack_range=1,
            attack_strength=1,
            attack_accuracy=1,
            view_range=3,
            **kwargs
        )


class TeamBattleSim(GridWorldSimulation):
    def __init__(self, **kwargs):
        self.agents = kwargs['agents']

        # State Components
        self.position_state = PositionState(**kwargs)
        self.health_state = HealthState(**kwargs)

        # Action Components
        self.move_actor = MoveActor(**kwargs)
        self.attack_actor = AttackActor(**kwargs)

        # Observation Components
        self.grid_observer = SingleGridObserver(**kwargs)

        # Done Compoennts
        self.done = OneTeamRemainingDone(**kwargs)

        self.finalize()

    def reset(self, **kwargs):
        self.position_state.reset(**kwargs)
        self.health_state.reset(**kwargs)

        # Track the rewards
        self.rewards = {agent.id: 0 for agent in self.agents.values()}

    def step(self, action_dict, **kwargs):
        # Process attacks:
        for agent_id, action in action_dict.items():
            agent = self.agents[agent_id]
            attacked_agent = self.attack_actor.process_action(agent, action, **kwargs)
            if attacked_agent is not None:
                self.rewards[attacked_agent.id] -= 1
                self.rewards[agent.id] += 1
            else:
                self.rewards[agent.id] -= 0.1

        # Process moves
        for agent_id, action in action_dict.items():
            agent = self.agents[agent_id]
            if agent.active:
                move_result = self.move_actor.process_action(agent, action, **kwargs)
                if not move_result:
                    self.rewards[agent.id] -= 0.1

        # Entropy penalty
        for agent_id in action_dict:
            self.rewards[agent_id] -= 0.01

    def get_obs(self, agent_id, **kwargs):
        agent = self.agents[agent_id]
        return {
            **self.grid_observer.get_obs(agent, **kwargs)
        }

    def get_reward(self, agent_id, **kwargs):
        reward = self.rewards[agent_id]
        self.rewards[agent_id] = 0
        return reward

    def get_done(self, agent_id, **kwargs):
        return self.done.get_done(self.agents[agent_id])

    def get_all_done(self, **kwargs):
        return self.done.get_all_done(**kwargs)

    def get_info(self, agent_id, **kwargs):
        return {}


if __name__ == "__main__":
    colors = ['red', 'blue', 'green', 'gray']
    positions = [np.array([1, 1]), np.array([1, 6]), np.array([6, 1]), np.array([6, 6])]
    agents = {
        f'agent{i}': BattleAgent(
            id=f'agent{i}',
            encoding=i % 4 + 1,
            render_color=colors[i % 4],
            initial_position=positions[i % 4]
        ) for i in range(24)
    }
    overlap_map = {
        1: [1],
        2: [2],
        3: [3],
        4: [4]
    }
    attack_map = {
        1: [2, 3, 4],
        2: [1, 3, 4],
        3: [1, 2, 4],
        4: [1, 2, 3]
    }
    sim = TeamBattleSim.build_sim(
        8, 8,
        agents=agents,
        overlapping=overlap_map,
        attack_mapping=attack_map
    )

    sim.reset()
    fig = plt.figure()
    sim.render(fig=fig)

    done_agents = set()
    for i in range(50):
        action = {
            agent.id: agent.action_space.sample()
            for agent in agents.values() if agent.id not in done_agents
        }
        sim.step(action)
        sim.render(fig=fig)

        if sim.get_all_done():
            break
        for agent in agents:
            if sim.get_done(agent):
                done_agents.add(agent)



from abc import ABC, abstractmethod

from ray.rllib.policy.policy import Policy


class HeuristicPolicy(Policy, ABC):
    """Abstract class for policies that do not learn."""
    @abstractmethod
    def compute_actions(self, *args, **kwargs):
        pass

    def learn_on_batch(self, samples):
        pass

    def get_weights(self):
        pass

    def set_weights(self, weights):
        pass



from abc import ABC, abstractmethod

import numpy as np


class Policy(ABC):
    """
    A policy maps a observation to an action. The relationship between the observations and the
    available actions is stored in a q_table. The act function chooses an action given a state.
    The probability function returns the probability of choosing an action given a state.
    """
    def __init__(self, q_table):
        """Store a q_table, which maps (state, action) to a value."""
        self.q_table = q_table

    @abstractmethod
    def act(self, state, *args, **kwargs):
        """Choose an action given a state."""
        pass

    @abstractmethod
    def probability(self, state, action):
        """Calculate the probability of choosing this action given this state."""
        pass

    def reset(self):
        """
        Some policies behave differently at the beginning of an episode or as an episode
        progresses. The reset function allows them to reset their parameters accordingly.
        """
        pass


class GreedyPolicy(Policy):
    """
    The GreedyPolicy will always choose the optimal action.
    """
    def act(self, state):
        return np.argmax(self.q_table[state])

    def probability(self, state, action):
        return 1 if action == np.argmax(self.q_table[state]) else 0


class EpsilonSoftPolicy(GreedyPolicy):
    """
    The EpsilonSoftPolicy will sample a uniform distribution between 0 and 1. If the sampled
    value is less than epsilon, then the policy will randomly choose an action. Otherwise, it
    will return the optimal action.
    """
    def __init__(self, *args, epsilon=0.1):
        super().__init__(*args)
        assert 0 <= epsilon <= 1.0
        self.epsilon = epsilon

    def act(self, state):
        if np.random.uniform(0, 1) < self.epsilon:
            return np.random.randint(0, self.q_table[state].size)
        else:
            return super().act(state)

    def probability(self, state, action):
        if action == np.argmax(self.q_table[state]): # Optimal action
            return 1 - self.epsilon + self.epsilon / self.q_table[state].size
        else: # Nonoptimal action
            return self.epsilon / self.q_table[state].size


class RandomFirstActionPolicy(GreedyPolicy):
    """
    The RandomFirstActionPolicy will choose a random action at the beginning of the episode.
    Afterwards, it will behave like a GreedyPolicy. Make sure you call the reset function at the
    beginning of every episode so that the policy knows to reset its parameters.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def reset(self):
        """
        Set take_random_action to True so that the policy takes a random action at the beginning
        of an episode.
        """
        self.take_random_action = True

    def act(self, state):
        if self.take_random_action:
            action = np.random.randint(0, self.q_table[state].size)
        else:
            action = super().act(state)
        self.take_random_action = False
        return action

    def probability(self, state, action):
        if self.take_random_action:
            return 1. / self.q_table[state].size
        else:
            return super().probability(state, action)




from abc import ABC

from matplotlib import pyplot as plt
import numpy as np

from abmarl.sim import AgentBasedSimulation
from abmarl.sim.gridworld.agent import GridWorldAgent
from abmarl.sim.gridworld.grid import Grid
from abmarl.tools.matplotlib_utils import mscatter


class GridWorldSimulation(AgentBasedSimulation, ABC):
    """
    GridWorldSimulation interface.

    Extends the AgentBasedSimulation interface for the GridWorld. We provide builders
    for streamlining the building process.
    """
    @classmethod
    def build_sim(cls, rows, cols, **kwargs):
        """
        Build a GridSimulation.

        Specify the number of row, the number of cols, a dictionary of agents,
        and any additional parameters.

        Args:
            rows: The number of rows in the grid. Must be a positive integer.
            cols: The number of cols in the grid. Must be a positive integer.
            agents: The dictionary of agents in the grid.

        Returns:
            A GridSimulation configured as specified.
        """
        assert type(rows) is int, "Rows must be an integer."
        assert 0 < rows, "Rows must be a positive integer."
        assert type(cols) is int, "Cols must be an integer."
        assert 0 < cols, "Cols must be a positive integer."

        return cls._build_sim(rows, cols, **kwargs)

    @classmethod
    def build_sim_from_file(cls, file_name, object_registry, **kwargs):
        """
        Build a GridSimulation from a text file.

        Args:
            file_name: Name of the file that specifies the initial grid setup. In the file, each
                cell should be a single alphanumeric character indicating which agent
                will be at that position (from the perspective of looking down on the
                grid). That agent will be given that initial position. 0's are reserved for empty
                space.
            object_registry: A dictionary that maps characters from the file to a
                function that generates the agent. This must be a function because
                each agent must have unique id, which is generated here.

        Returns:
            A GridSimulation built from the file.
        """
        assert type(file_name) is str, "The file_name must be the name of the file."
        assert type(object_registry) is dict, "The object_registry must be a dictionary."
        assert 0 not in object_registry, "0 is reserved for empty space."
        agents = {}
        n = 0
        with open(file_name, 'r') as fp:
            lines = fp.read().splitlines()
            cols = len(lines[0].split(' '))
            rows = len(lines)
            for row, line in enumerate(lines):
                chars = line.split(' ')
                assert len(chars) == cols, f"Mismatched number of columns per row in {file_name}"
                for col, char in enumerate(chars):
                    if char in object_registry:
                        agent = object_registry[char](n)
                        agent.initial_position = np.array([row, col])
                        agents[agent.id] = agent
                        n += 1

        return cls._build_sim(rows, cols, agents=agents, **kwargs)

    @classmethod
    def _build_sim(cls, rows, cols, **kwargs):
        grid = Grid(rows, cols, **kwargs)
        kwargs['grid'] = grid
        return cls(**kwargs)

    def render(self, fig=None, **kwargs):
        """
        Draw the grid and all active agents in the grid.

        Agents are drawn at their positions using their respective shape and color.

        Args:
            fig: The figure on which to draw the grid. It's important
                to provide this figure because the same figure must be used when drawing
                each state of the simulation. Otherwise, a ton of figures will pop up,
                which is very annoying.
        """
        fig.clear()
        ax = fig.gca()

        # Draw the gridlines
        ax.set(xlim=(0, self.position_state.cols), ylim=(0, self.position_state.rows))
        ax.set_xticks(np.arange(0, self.position_state.cols, 1))
        ax.set_yticks(np.arange(0, self.position_state.rows, 1))
        ax.grid()

        # Draw the agents
        agents_x = [
            agent.position[1] + 0.5 for agent in self.agents.values() if agent.active
        ]
        agents_y = [
            self.position_state.rows - 0.5 - agent.position[0]
            for agent in self.agents.values() if agent.active
        ]
        shape = [agent.render_shape for agent in self.agents.values() if agent.active]
        color = [agent.render_color for agent in self.agents.values() if agent.active]
        mscatter(agents_x, agents_y, ax=ax, m=shape, s=200, facecolor=color)

        plt.plot()
        plt.pause(1e-6)


class GridWorldBaseComponent(ABC):
    """
    Component base class from which all components will inherit.

    Every component has access to the dictionary of agents and the grid.
    """
    def __init__(self, agents=None, grid=None, **kwargs):
        self.agents = agents
        self.grid = grid

    @property
    def rows(self):
        """
        The number of rows in the grid.
        """
        return self.grid.rows

    @property
    def cols(self):
        """
        The number of columns in the grid.
        """
        return self.grid.cols

    @property
    def grid(self):
        """
        The grid indexes the agents by their position.

        For example, an agent whose position is (3, 2) can be accessed through
        the grid with ``self.grid[3, 2]``. Components are responsible for maintaining
        the connection between agent position and grid index.
        """
        return self._grid

    @grid.setter
    def grid(self, value):
        assert isinstance(value, Grid), "The grid must be a Grid object."
        self._grid = value

    @property
    def agents(self):
        """
        A dict that maps the Agent's id to the Agent object. All agents must be
        GridWorldAgents.
        """
        return self._agents

    @agents.setter
    def agents(self, value_agents):
        assert type(value_agents) is dict, "Agents must be a dict."
        for agent_id, agent in value_agents.items():
            assert isinstance(agent, GridWorldAgent), \
                "Values of agents dict must be instance of GridWorldAgent."
            assert agent_id == agent.id, \
                "Keys of agents dict must be the same as the Agent's id."
        self._agents = value_agents



from .abstract_policy import HeuristicPolicy
from .random_policy import RandomAction

from .policy import GreedyPolicy
from .policy import EpsilonSoftPolicy
from .policy import RandomFirstActionPolicy



from gym.spaces import Discrete
import numpy as np

from abmarl.sim.wrappers import CommunicationHandshakeWrapper

from .helpers import MultiAgentGymSpacesSim


class CommsSimulation(MultiAgentGymSpacesSim):
    def get_obs(self, my_id, fusion_matrix={}):
        my_obs = super().get_obs(my_id)

        if 'agent1' in fusion_matrix and fusion_matrix['agent1']:
            agent_1_obs = self.get_obs('agent1')
            if my_id == 'agent0':
                my_obs[0] += agent_1_obs
            elif my_id == 'agent3':
                my_obs['first'] = agent_1_obs
        elif 'agent2' in fusion_matrix and fusion_matrix['agent2']:
            agent_2_obs = self.get_obs('agent2')
            if my_id == 'agent0':
                my_obs[1:2] += agent_2_obs[:]
            elif my_id == 'agent3':
                my_obs['second'] = agent_2_obs
        elif 'agent3' in fusion_matrix and fusion_matrix['agent3']:
            agent_3_obs = self.get_obs('agent3')
            if my_id == 'agent0':
                my_obs[0:-1] += [agent_3_obs['first'], agent_3_obs['second']][:]

        return my_obs


def test_communication_wrapper_init():
    sim = CommsSimulation()
    wrapped_sim = CommunicationHandshakeWrapper(sim)
    assert wrapped_sim.sim == sim
    assert wrapped_sim.agents != sim.agents
    assert wrapped_sim.unwrapped == sim

    for agent_id, agent in wrapped_sim.agents.items():
        assert agent.action_space['action'] == sim.agents[agent_id].action_space
        assert agent.observation_space['obs'] == sim.agents[agent_id].observation_space

        for other_agent in wrapped_sim.agents:
            if other_agent == agent_id: continue
            assert agent.action_space['send'][other_agent] == Discrete(2)
            assert agent.action_space['receive'][other_agent] == Discrete(2)
            assert agent.observation_space['message_buffer'][other_agent] == Discrete(2)


def test_communication_wrapper_reset():
    sim = CommunicationHandshakeWrapper(CommsSimulation())
    sim.reset()
    for values in sim.message_buffer.values():
        assert all([True if not val else False for val in values.values()])
    for values in sim.received_message.values():
        assert all([True if not val else False for val in values.values()])

    for agent_id in sim.agents:
        assert 'obs' in sim.get_obs(agent_id)
    assert sim.message_buffer == {
        agent_id: sim.get_obs(agent_id)['message_buffer'] for agent_id in sim.agents
    }


def test_communication_wrapper_step():
    sim = CommunicationHandshakeWrapper(CommsSimulation())
    sim.reset()

    action_0 = {
        'agent0': {
            'action': ({'first': 2, 'second': [-1, 2]}, [0, 1, 1]),
            'send': {'agent1': True, 'agent2': True, 'agent3': True},
            'receive': {'agent1': True, 'agent2': True, 'agent3': True},
        },
        'agent1': {
            'action': [3, 2, 0],
            'send': {'agent0': True, 'agent2': False, 'agent3': False},
            'receive': {'agent0': True, 'agent2': True, 'agent3': True},
        },
        'agent2': {
            'action': {'alpha': [1, 1, 0]},
            'send': {'agent0': True, 'agent1': True, 'agent3': False},
            'receive': {'agent0': True, 'agent1': True, 'agent3': True},
        },
        'agent3': {
            'action': (2, np.array([0, 6]), 1),
            'send': {'agent0': False, 'agent1': False, 'agent2': True},
            'receive': {'agent0': True, 'agent1': True, 'agent2': True},
        }
    }
    sim.step(action_0)
    for agent_id in sim.agents:
        agent_info = sim.get_info(agent_id)
        assert 'send' not in agent_info and 'receive' not in agent_info
    assert sim.get_obs('agent0')['obs'] == [0, 0, 0, 1]
    assert sim.get_obs('agent0')['message_buffer'] == {
        'agent1': True, 'agent2': True, 'agent3': False
    }
    assert sim.get_obs('agent1')['obs'] == 0
    assert sim.get_obs('agent1')['message_buffer'] == {
        'agent0': True, 'agent2': True, 'agent3': False
    }
    assert sim.get_obs('agent2')['obs'] == [1, 0]
    assert sim.get_obs('agent2')['message_buffer'] == {
        'agent0': True, 'agent1': False, 'agent3': True
    }
    assert sim.get_obs('agent3')['obs'] == {'first': 1, 'second': [3, 1]}
    assert sim.get_obs('agent3')['message_buffer'] == {
        'agent0': True, 'agent1': False, 'agent2': False
    }

    action_1 = {
        'agent0': {
            'action': ({'first': 0, 'second': [3, 3]}, [1, 1, 1]),
            'send': {'agent1': False, 'agent2': False, 'agent3': False},
            'receive': {'agent1': True, 'agent2': False, 'agent3': False},
        },
        'agent1': {
            'action': [1, 5, 1],
            'send': {'agent0': True, 'agent2': False, 'agent3': True},
            'receive': {'agent0': False, 'agent2': False, 'agent3': True},
        },
        'agent2': {
            'action': {'alpha': [1, 0, 0]},
            'send': {'agent0': True, 'agent1': True, 'agent3': False},
            'receive': {'agent0': True, 'agent1': True, 'agent3': False},
        },
        'agent3': {
            'action': (1, np.array([9, 4]), 0),
            'send': {'agent0': False, 'agent1': True, 'agent2': True},
            'receive': {'agent0': True, 'agent1': False, 'agent2': True},
        }
    }
    sim.step(action_1)
    assert sim.get_obs('agent0')['obs'] == [0, 0, 0, 1]
    assert sim.get_obs('agent0')['message_buffer'] == {
        'agent1': True, 'agent2': True, 'agent3': False
    }
    assert sim.get_obs('agent1')['obs'] == 0
    assert sim.get_obs('agent1')['message_buffer'] == {
        'agent0': False, 'agent2': True, 'agent3': True
    }
    assert sim.get_obs('agent2')['obs'] == [1, 0]
    assert sim.get_obs('agent2')['message_buffer'] == {
        'agent0': False, 'agent1': False, 'agent3': True
    }
    assert sim.get_obs('agent3')['obs'] == {'first': 1, 'second': [3, 1]}
    assert sim.get_obs('agent3')['message_buffer'] == {
        'agent0': False, 'agent1': True, 'agent2': False
    }

    action_2 = {
        'agent0': {
            'action': ({'first': 1, 'second': [1, 0]}, [0, 0, 1]),
            'send': {'agent1': True, 'agent2': True, 'agent3': True},
            'receive': {'agent1': False, 'agent2': True, 'agent3': True},
        },
        'agent1': {
            'action': [2, 0, 1],
            'send': {'agent0': True, 'agent2': False, 'agent3': False},
            'receive': {'agent0': True, 'agent2': True, 'agent3': True},
        },
        'agent2': {
            'action': {'alpha': [0, 0, 0]},
            'send': {'agent0': True, 'agent1': True, 'agent3': False},
            'receive': {'agent0': True, 'agent1': True, 'agent3': False},
        },
        'agent3': {
            'action': (0, np.array([7, 7]), 0),
            'send': {'agent0': False, 'agent1': False, 'agent2': True},
            'receive': {'agent0': True, 'agent1': True, 'agent2': False},
        }
    }
    sim.step(action_2)
    assert sim.get_obs('agent0')['obs'] == [0, 0, 1, 0, 0, 1]
    assert sim.get_obs('agent0')['message_buffer'] == {
        'agent1': True, 'agent2': True, 'agent3': False
    }
    assert sim.get_obs('agent1')['obs'] == 0
    assert sim.get_obs('agent1')['message_buffer'] == {
        'agent0': True, 'agent2': True, 'agent3': False
    }
    assert sim.get_obs('agent2')['obs'] == [1, 0]
    assert sim.get_obs('agent2')['message_buffer'] == {
        'agent0': True, 'agent1': False, 'agent3': True
    }
    assert sim.get_obs('agent3')['obs'] == {'first': 0, 'second': [3, 1]}
    assert sim.get_obs('agent3')['message_buffer'] == {
        'agent0': True, 'agent1': False, 'agent2': False
    }



def mscatter(x, y, ax=None, m=None, **kw):
    """
    Scatter plot with markers m assigned to each element.
    """
    import matplotlib.pyplot as plt
    import matplotlib.markers as mmarkers
    ax = ax or plt.gca()
    sc = ax.scatter(x, y, **kw)
    if (m is not None) and (len(m) == len(x)):
        paths = []
        for marker in m:
            if isinstance(marker, mmarkers.MarkerStyle):
                marker_obj = marker
            else:
                marker_obj = mmarkers.MarkerStyle(marker)
            path = marker_obj.get_path().transformed(marker_obj.get_transform())
            paths.append(path)
        sc.set_paths(paths)
    return sc



import pytest

from abmarl.managers import AllStepManager
from abmarl.external import GymWrapper

from .helpers import MultiAgentSim


def test_gym_init_multi_agent_error():
    with pytest.raises(AssertionError):
        GymWrapper(MultiAgentSim(3))


def test_gym_init():
    sim = AllStepManager(MultiAgentSim(1))
    wrapped_sim = GymWrapper(sim)
    assert wrapped_sim.sim == sim


def test_gym_reset_and_step():
    sim = GymWrapper(AllStepManager(MultiAgentSim(1)))
    obs = sim.reset()
    assert obs == 'Obs from agent0'
    obs, reward, done, info = sim.step(0)
    assert obs == 'Obs from agent0'
    assert reward == 'Reward from agent0'
    assert done == 'Done from agent0'
    assert info == {'Action from agent0': 0}



from itertools import cycle

from abmarl.sim import ActingAgent, ObservingAgent

from .simulation_manager import SimulationManager


class TurnBasedManager(SimulationManager):
    """
    The TurnBasedManager allows agents to take turns. The order of the agents is stored and the
    obs of the first agent is returned at reset. Each step returns the info of
    the next agent "in line". Agents who are done are removed from this line.
    Once all the agents are done, the manager returns all done.
    """
    def __init__(self, sim):
        super().__init__(sim)
        self.agent_order = cycle({
            agent_id: agent for agent_id, agent in self.agents.items()
            if (isinstance(agent, ActingAgent) and isinstance(agent, ObservingAgent))
        })

    def reset(self, **kwargs):
        """
        Reset the simulation and return the observation of the first agent.
        """
        self.done_agents = set()

        self.sim.reset(**kwargs)
        next_agent = next(self.agent_order)
        return {next_agent: self.sim.get_obs(next_agent)}

    def step(self, action_dict, **kwargs):
        """
        Assert that the incoming action does not come from an agent who is recorded
        as done. Step the simulation forward and return the observation, reward,
        done, and info of the next agent. If that next agent finished in this turn,
        then include the obs for the following agent, and so on until an agent
        is found that is not done. If all agents are done in this turn, then the
        wrapper returns all done.
        """
        agent_id = next(iter(action_dict))
        assert agent_id not in self.done_agents, \
            "Received an action for an agent that is already done."
        self.sim.step(action_dict, **kwargs)

        obs, rewards, dones, infos = {}, {}, {'__all__': self.sim.get_all_done()}, {}
        if dones['__all__']: # The simulation is done. Get output for all non-done agents
            for agent in self.agents:
                if agent in self.done_agents:
                    continue
                else:
                    obs[agent] = self.sim.get_obs(agent)
                    rewards[agent] = self.sim.get_reward(agent)
                    dones[agent] = self.sim.get_done(agent)
                    infos[agent] = self.sim.get_info(agent)
        else: # Simulation is not done. Get the output for the next agent(s).
            for next_agent in self.agent_order:
                # This agent was already done before, so there is no interaction
                # with it
                if next_agent in self.done_agents: continue

                # Check if the agent is just recently done:
                elif self.sim.get_done(next_agent):
                    # This agent only just recently finished. It sent an action before
                    # and now expects to receive an observation, rewrard, and done signal.
                    # So I want to add that to the output, but I don't want its action
                    # because it is done. So I want to include its info AND the info from
                    # the next not done agent.
                    obs[next_agent] = self.sim.get_obs(next_agent)
                    rewards[next_agent] = self.sim.get_reward(next_agent)
                    dones[next_agent] = self.sim.get_done(next_agent)
                    infos[next_agent] = self.sim.get_info(next_agent)
                    self.done_agents.add(next_agent)

                    # All agents could potentially be done now, so we check for that
                    if any([True for agent in self.agents if agent not in self.done_agents]):
                        continue
                    else:
                        # All agents are done
                        dones['__all__'] = True
                        break

                else:
                    # The agent is not done at all. So we grab its information and
                    # break the agent iteration loop
                    obs[next_agent] = self.sim.get_obs(next_agent)
                    rewards[next_agent] = self.sim.get_reward(next_agent)
                    dones[next_agent] = self.sim.get_done(next_agent)
                    infos[next_agent] = self.sim.get_info(next_agent)
                    break

        return obs, rewards, dones, infos



import numpy as np

from abmarl.managers import TurnBasedManager, AllStepManager
from abmarl.sim.corridor import MultiCorridor as Corridor


def test_managers_are_same_for_single_agent():
    turn_based_sim = TurnBasedManager(Corridor(num_agents=1))
    all_step_sim = AllStepManager(Corridor(num_agents=1))

    np.random.seed(5)
    turn_based_reset = turn_based_sim.reset()
    np.random.seed(5)
    all_step_reset = all_step_sim.reset()
    assert turn_based_reset == all_step_reset

    assert turn_based_sim.step({'agent0': 0}) == all_step_sim.step({'agent0': 0})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})
    assert turn_based_sim.step({'agent0': 1}) == all_step_sim.step({'agent0': 1})
    assert turn_based_sim.step({'agent0': 0}) == all_step_sim.step({'agent0': 0})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})
    assert turn_based_sim.step({'agent0': 2}) == all_step_sim.step({'agent0': 2})



def create_parser(subparsers):
    """Parse the arguments for the make_runnable command.

    Returns
    -------
        parser : ArgumentParser
    """
    runnable_parser = subparsers.add_parser(
        'make-runnable',
        help='Convert a configuration + script to be runnable from the commandline. Save the '
        'coverted script in the same directory as the original script. This option is useful for '
        'running batch jobs from the command line and/or integration with magpie.'
    )
    runnable_parser.add_argument(
        'configuration', type=str, help='Path to python config file. Include the .py extension.'
    )
    runnable_parser.add_argument(
        '--magpie', action='store_true', help='Output a magpie script as well.'
    )
    runnable_parser.add_argument(
        '-n', '--nodes', type=int,
        help='Specify the number of compute nodes (not rollout workers) you want.', default=2
    )
    runnable_parser.add_argument(
        '-t', '--time-limit', type=str, help='The maximum runtime for this job in hours',
        default='2'
    )
    return runnable_parser


def run(full_config_path, parameters):
    from abmarl import make_runnable
    make_runnable.run(full_config_path, parameters)



from .wrapper import Wrapper

from gym.spaces import Discrete, Dict


class CommunicationHandshakeWrapper(Wrapper):
    """
    Agents can share their observations with one another according to this communication
    protocol. We allow agents to send and receive messages from any other agent
    in the simulation. When an agent chooses to send, recipient agents will see in their next
    observation that there are incoming messages. Agents can choose to receive
    or ignore the message. If an agent chooses to receive, its observation space
    is fused with the sending agent's observation. That fusion happens according
    to the fuse_observation function, which must be implemented by the wrapped simulation.

    Agents' observation and action spaces are converted into dictionaries. The
    action and observation from the simulation are keyed on 'action' and 'obs',
    respectively. We add 'message_buffer' to the observation, which shows incoming
    messages; and 'receive' and 'send' to the action, which are the two communication
    actions the agents can take.
    """
    def __init__(self, sim):
        super().__init__(sim)

        # Augment the agents' action and observation spaces.
        # We use a dict keyed off the agents' id to Discrete(2) instead of just
        # MultiBinary(num_agents-1) because MultiBinary only gives us
        # T/F at some indicies. We would need additional mapping information to
        # map from the index of the MultiBinary observation/action to the respective
        # agent. Using a dict gives us that information automatically.
        for agent in self.agents.values():
            action_space_helper = {'action': agent.action_space}
            action_space_helper['send'] = Dict({
                other_id: Discrete(2) for other_id in self.agents if other_id != agent.id
            })
            action_space_helper['receive'] = Dict({
                other_id: Discrete(2) for other_id in self.agents if other_id != agent.id
            })
            agent.action_space = Dict(action_space_helper)

            obs_space_helper = {'obs': agent.observation_space}
            obs_space_helper['message_buffer'] = Dict({
                other_id: Discrete(2) for other_id in self.agents if other_id != agent.id
            })
            agent.observation_space = Dict(obs_space_helper)

    def reset(self, **kwargs):
        """
        Set the internal communication state to the null state and reset the wrapped
        simulation.
        """
        self.message_buffer = {}
        self.received_message = {}
        for my_id in self.agents:
            self.message_buffer[my_id] = {
                other_id: False for other_id in self.agents if other_id != my_id
            }
            self.received_message[my_id] = {
                other_id: False for other_id in self.agents if other_id != my_id
            }
        self.sim.reset(**kwargs)

    def step(self, action_dict, **kwargs):
        """
        First, we process the receive actions, which will be used to determine
        any message that are successfully communicated among agents. Then, we call
        the step function from the wrapped simulation. Finally, we process the
        send actions to update the message buffer observations.
        """
        # Process receive actions
        for receiving_agent, action in action_dict.items():
            self.received_message[receiving_agent] = {
                sending_agent: True if self.message_buffer[receiving_agent][sending_agent] and
                action['receive'][sending_agent] else False
                for sending_agent in self.received_message[receiving_agent]
            }
        # Reset the message buffer
        for my_id in self.agents:
            self.message_buffer[my_id] = {
                other_id: False for other_id in self.agents if other_id != my_id
            }

        # Wrapped simulation takes a step
        sim_only_action = {
            agent_id: action_dict[agent_id]['action'] for agent_id in action_dict
        }
        self.sim.step(sim_only_action, **kwargs)

        # Process send actions
        for sending_agent, action in action_dict.items():
            for receiving_agent, message in action['send'].items():
                self.message_buffer[receiving_agent][sending_agent] = message

    def get_obs(self, agent_id, **kwargs):
        """
        The (fused) observation from the wrapped simulation is keyed on 'obs'
        and we add 'message_buffer' for incoming messages.
        """
        obs_from_sim = self.sim.get_obs(agent_id, fusion_matrix=self.received_message[agent_id])
        return {'obs': obs_from_sim, 'message_buffer': self.message_buffer[agent_id]}



# 5.1 Monte Carlo prediction p92.
from gym.spaces import Discrete
import numpy as np

from abmarl.managers import SimulationManager
from abmarl.external import GymWrapper
from abmarl.pols import GreedyPolicy, EpsilonSoftPolicy, RandomFirstActionPolicy
from abmarl.tools import numpy_utils as npu

from .generate_episode import generate_episode


def exploring_starts(sim, iteration=10_000, gamma=0.9, horizon=200):
    """
    Estimate an optimal policy over an simulation using monte carlo policy estimation.

    Args:
        sim: The simulation, obviously.
        iteration: The number of times to iterate the learning algorithm.
        gamma: The discount factor
        horizon: the time horizon for the trajectory.

    Returns:
        sim: The simulation. Algorithms may wrap simulations before training in them, so this
            simulation may be wrapped.
        q_table: The Q values
        policy: The policy that is learned.
    """
    assert isinstance(sim, SimulationManager)
    sim = GymWrapper(sim)
    assert isinstance(sim.observation_space, Discrete)
    assert isinstance(sim.action_space, Discrete)
    q_table = np.random.normal(0, 1, size=(sim.observation_space.n, sim.action_space.n))
    policy = RandomFirstActionPolicy(q_table)
    state_action_returns = {}

    for i in range(iteration):
        states, actions, rewards = generate_episode(sim, policy, horizon)
        states = np.stack(states)
        actions = np.stack(actions)
        G = 0
        for i in reversed(range(len(states))):
            state, action, reward = states[i], actions[i], rewards[i]
            G = gamma * G + reward
            if not (npu.array_in_array(state, states[:i]) and
                    npu.array_in_array(action, actions[:i])):
                if (state, action) not in state_action_returns:
                    state_action_returns[(state, action)] = [G]
                else:
                    state_action_returns[(state, action)].append(G)
                q_table[state, action] = np.mean(state_action_returns[(state, action)])

    return sim, q_table, policy


def epsilon_soft(sim, iteration=10_000, gamma=0.9, epsilon=0.1, horizon=200):
    """
    Estimate an optimal policy over a simulation using monte carlo policy estimation. The policy
    is technically non-optimal because it is epsilon-soft.

    Args:
        sim: The simulation, obviously.
        iteration: The number of times to iterate the learning algorithm.
        gamme: The discount factor
        epsilon: The exploration probability.

    Returns:
        sim: The simulation. Algorithms may wrap simulations before training in them, so this
            simulation may be wrapped.
        q_table: The Q values
        policy: The policy that is learned.
    """
    assert isinstance(sim, SimulationManager)
    sim = GymWrapper(sim)
    assert isinstance(sim.observation_space, Discrete)
    assert isinstance(sim.action_space, Discrete)
    q_table = np.random.normal(0, 1, size=(sim.observation_space.n, sim.action_space.n))
    policy = EpsilonSoftPolicy(q_table, epsilon=epsilon)
    state_action_returns = {}

    for i in range(iteration):
        states, actions, rewards = generate_episode(sim, policy, horizon)
        states = np.stack(states)
        actions = np.stack(actions)
        G = 0
        for i in reversed(range(len(states))):
            state, action, reward = states[i], actions[i], rewards[i]
            G = gamma * G + reward
            if not (npu.array_in_array(state, states[:i]) and
                    npu.array_in_array(action, actions[:i])):
                if (state, action) not in state_action_returns:
                    state_action_returns[(state, action)] = [G]
                else:
                    state_action_returns[(state, action)].append(G)
                q_table[state, action] = np.mean(state_action_returns[(state, action)])

    return sim, q_table, policy


def off_policy(sim, iteration=10_000, gamma=0.9, horizon=200):
    """
    Off-policy Monte Carlo control estimates an optimal policy in a simulation. Trains a greedy
    policy be generating trajectories from an epsilon-soft behavior policy.

    Args:
        sim: The simulation, obviously.
        iteration: The number of times to iterate the learning algorithm.
        gamme: The discount factor

    Returns:
        sim: The simulation. Algorithms may wrap simulations before training in them, so this
            simulation may be wrapped.
        q_table: The Q values
        policy: The policy that is learned.
    """
    assert isinstance(sim, SimulationManager)
    sim = GymWrapper(sim)
    assert isinstance(sim.observation_space, Discrete)
    assert isinstance(sim.action_space, Discrete)
    q_table = np.random.normal(0, 1, size=(sim.observation_space.n, sim.action_space.n))
    c_table = 0 * q_table
    policy = GreedyPolicy(q_table)
    for i in range(iteration):
        behavior_policy = EpsilonSoftPolicy(q_table)
        states, actions, rewards, = generate_episode(sim, behavior_policy, horizon)
        G = 0
        W = 1
        for i in reversed(range(len(states))):
            state, action, reward = states[i], actions[i], rewards[i]
            G = gamma * G + reward
            c_table[state, action] += W
            q_table[state, action] = q_table[state, action] + W/(c_table[state, action]) * \
                (G - q_table[state, action])
            if action != policy.act(state): # Nonoptimal action
                break
            W /= behavior_policy.probability(state, action)

    return sim, q_table, policy



from abc import ABC, abstractmethod

from abmarl.tools import gym_utils as gu


class PrincipleAgent:
    """
    Principle Agent class for agents in a simulation.
    """
    def __init__(self, id=None, seed=None, **kwargs):
        self.id = id
        self.seed = seed

    @property
    def id(self):
        """The agent's unique identifier."""
        return self._id

    @id.setter
    def id(self, value):
        assert type(value) is str, "id must be a string."
        self._id = value

    @property
    def seed(self):
        """Seed for random number generation."""
        return self._seed

    @seed.setter
    def seed(self, value):
        assert value is None or type(value) is int, "Seed must be an integer."
        self._seed = value

    @property
    def active(self):
        """
        True if the agent is still active in the simulation.

        Active means that the agent is in a valid state. For example, suppose agents
        in our Simulation can die. Then active is True if the agents are alive
        or False if they're dead.
        """
        return True

    @property
    def configured(self):
        """
        All agents must have an id.
        """
        return self.id is not None

    def finalize(self, **kwargs):
        pass

    def __eq__(self, other):
        return self.__dict__ == other.__dict__ if isinstance(other, self.__class__) else False


class ActingAgent(PrincipleAgent):
    """
    ActingAgents can act in the simulation.

    The Trainer will produce actions for the agents and send them to the SimulationManager,
    which will process those actions in its step function.
    """
    def __init__(self, action_space=None, **kwargs):
        super().__init__(**kwargs)
        self.action_space = action_space

    @property
    def action_space(self):
        return self._action_space

    @action_space.setter
    def action_space(self, value):
        assert value is None or gu.check_space(value), \
            "The action space must be None, a gym Space, or a dict of gym Spaces."
        self._action_space = {} if value is None else value

    @property
    def configured(self):
        """
        Acting agents must have an action space.
        """
        return super().configured and gu.check_space(self.action_space, strict=True)

    def finalize(self, **kwargs):
        """
        Wrap all the action spaces with a Dict if applicable and seed it if the agent was
        created with a seed.
        """
        super().finalize(**kwargs)
        if type(self.action_space) is dict:
            self.action_space = gu.make_dict(self.action_space)
        self.action_space.seed(self.seed)


class ObservingAgent(PrincipleAgent):
    """
    ObservingAgents can observe the state of the simulation.

    The agent's observation must be *in* its observation space. The SimulationManager
    will send the observation to the Trainer, which will use it to produce actions.
    """
    def __init__(self, observation_space=None, **kwargs):
        super().__init__(**kwargs)
        self.observation_space = observation_space

    @property
    def observation_space(self):
        return self._observation_space

    @observation_space.setter
    def observation_space(self, value):
        assert value is None or gu.check_space(value), \
            "The observation space must be None, a gym Space, or a dict of gym Spaces."
        self._observation_space = {} if value is None else value

    @property
    def configured(self):
        """
        Observing agents must have an observation space.
        """
        return super().configured and gu.check_space(self.observation_space, strict=True)

    def finalize(self, **kwargs):
        """
        Wrap all the observation spaces with a Dict and seed it if the agent was
        created with a seed.
        """
        super().finalize(**kwargs)
        if type(self.observation_space) is dict:
            self.observation_space = gu.make_dict(self.observation_space)
        self.observation_space.seed(self.seed)


class Agent(ObservingAgent, ActingAgent):
    """
    An Agent that can both observe and act.
    """
    pass


class AgentBasedSimulation(ABC):
    """
    AgentBasedSimulation interface.

    Under this design model the observations, rewards, and done conditions of the
    agents is treated as part of the simulations internal state instead of as
    output from reset and step. Thus, it is the simulations responsibility to manage
    rewards and dones as part of its state (e.g. via self.rewards dictionary).

    This interface supports both single- and multi-agent simulations by treating
    the single-agent simulation as a special case of the multi-agent, where there
    is only a single agent in the agents dictionary.
    """

    @property
    def agents(self):
        """
        A dict that maps the Agent's id to the Agent object. An Agent must be an
        instance of PrincipleAgent. A multi-agent simulation is expected to have
        multiple entries in the dictionary, whereas a single-agent simulation
        should only have a single entry in the dictionary.
        """

        return self._agents

    @agents.setter
    def agents(self, value_agents):
        assert type(value_agents) is dict, "Agents must be a dict."
        for agent_id, agent in value_agents.items():
            assert isinstance(agent, PrincipleAgent), \
                "Values of agents dict must be instance of PrincipleAgent."
            assert agent_id == agent.id, \
                "Keys of agents dict must be the same as the Agent's id."
        self._agents = value_agents

    def finalize(self):
        """
        Finalize the initialization process. At this point, every agent should
        be configured with action and observation spaces, which we convert into
        Dict spaces for interfacing with the trainer.
        """
        for agent in self.agents.values():
            agent.finalize()
            assert agent.configured

    @abstractmethod
    def reset(self, **kwargs):
        """
        Reset the simulation simulation to a start state, which may be randomly
        generated.
        """
        pass

    @abstractmethod
    def step(self, action, **kwargs):
        """
        Step the simulation forward one discrete time-step. The action is a dictionary
        that contains the action of each agent in this time-step.
        """
        pass

    @abstractmethod
    def render(self, **kwargs):
        """
        Render the simulation for vizualization.
        """
        pass

    @abstractmethod
    def get_obs(self, agent_id, **kwargs):
        """
        Return the agent's observation.
        """
        pass

    @abstractmethod
    def get_reward(self, agent_id, **kwargs):
        """
        Return the agent's reward.
        """
        pass

    @abstractmethod
    def get_done(self, agent_id, **kwargs):
        """
        Return the agent's done status.
        """
        pass

    @abstractmethod
    def get_all_done(self, **kwargs):
        """
        Return the simulation's done status.
        """
        pass

    @abstractmethod
    def get_info(self, agent_id, **kwargs):
        """
        Return the agent's info.
        """
        pass



from gym.spaces import Dict, Tuple, Box, Discrete, MultiDiscrete, MultiBinary
import numpy as np

from abmarl.sim import Agent
from abmarl.sim.wrappers import FlattenWrapper
from abmarl.sim.wrappers.flatten_wrapper import flatdim, flatten, unflatten, flatten_space
from .helpers import MultiAgentSim

# --- Test flatten helper commands --- #

box = Box(2, 16, (3,4), np.int)
box2 = Box(2.4, 16.1, (3,4))
discrete = Discrete(11)
multi_binary = MultiBinary(7)
multi_discrete = MultiDiscrete([2, 6, 4])
d = Dict({'first': box2, 'second': multi_binary})
t = Tuple((discrete, box2, multi_discrete))
combo = Tuple((Dict({'first': discrete, 'second': box}), multi_binary))


def test_integer_sample():
    flat_combo = flatten_space(combo)
    samp = flat_combo.sample()
    for i in samp:
        assert type(i) == np.int64


def test_flatdim():
    assert flatdim(box) == 12
    assert flatdim(discrete) == 11
    assert flatdim(multi_binary) == 7
    assert flatdim(multi_discrete) == 3
    assert flatdim(d) == 19
    assert flatdim(t) == 26
    assert flatdim(combo) == 30


def test_flatten_and_unflatten():
    box_sample = np.array([
        [2, 12, 6, 6],
        [2, 8, 5, 2],
        [2, 16, 4, 10]
    ])
    flattened_box_sample = np.array([
        2, 12, 6, 6, 2, 8, 5, 2, 2, 16, 4, 10
    ])

    box2_sample = np.array([
        [1.570356,   8.745239,  15.145958,   6.7516246],
        [ 4.981189,   8.553591,   4.252773,  13.892432 ],
        [ 9.568261,  13.409188,   3.5828538,  3.1473527]
    ])
    flattened_box2_sample = np.array([
        1.570356, 8.745239, 15.145958, 6.7516246, 4.981189, 8.553591, 4.252773,
        13.892432, 9.568261, 13.409188, 3.5828538, 3.1473527
    ])

    d_sample = {
        'first': np.array([
            [6.7937455, 14.242623 ,  8.434648 ,  3.721371 ],
            [13.182184 ,  2.9882407, 11.972553 , 10.094532 ],
            [ 7.314138 ,  8.059702 , 13.132411 ,  6.1159306]]
        ),
        'second': np.array([1, 1, 0, 0, 0, 0, 1])
    }
    flattened_d_sample = np.array([
        6.79374552, 14.24262333, 8.43464756, 3.72137094, 13.18218422, 2.98824072,
        11.97255325, 10.09453201, 7.31413794, 8.05970192, 13.132411, 6.11593056,
        1., 1., 0., 0., 0., 0., 1.
    ])

    t_sample = (
        6,
        np.array([
            [2.8780882, 9.45528, 7.886346, 5.2954407],
            [15.226127, 11.98857, 3.2817354, 7.7292967],
            [12.403905, 8.250312, 10.714847, 9.248549 ]
        ]),
        np.array([1, 1, 3])
    )
    flattened_t_sample = np.array([
        0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 2.87808824, 9.4552803, 7.88634586,
        5.29544067, 15.22612667, 11.98857021, 3.28173542, 7.72929668, 12.40390491,
        8.25031185, 10.71484661, 9.24854946, 1., 1., 3.
    ])

    combo_sample = (
        {
            'first': 2,
            'second': np.array([
                [15, 8, 10, 3],
                [14, 10, 7, 7],
                [7, 14, 4, 10]
            ])
        },
        np.array([1, 1, 1, 0, 1, 1, 0])
    )
    flattened_combo_sample = np.array([
        0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 15, 8, 10, 3, 14, 10, 7, 7, 7, 14, 4,
        10, 1, 1, 1, 0, 1, 1, 0
    ])

    # Test flattens
    np.testing.assert_array_equal(flatten(box, box_sample), flattened_box_sample)
    assert np.allclose(flatten(box2, box2_sample), flattened_box2_sample, atol=1.e-6)
    np.testing.assert_array_equal(
        flatten(discrete, 8), np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0])
    )
    np.testing.assert_array_equal(
        flatten(multi_binary, [0, 1, 0, 0, 1, 1, 1]), np.array([0, 1, 0, 0, 1, 1, 1])
    )
    np.testing.assert_array_equal(flatten(multi_discrete, [0, 3, 1]), np.array([0, 3, 1]))
    assert np.allclose(flatten(d, d_sample), flattened_d_sample, atol=1.e-8)
    assert np.allclose(flatten(t, t_sample), flattened_t_sample, atol=1.e-8)
    np.testing.assert_array_equal(flatten(combo, combo_sample), flattened_combo_sample)

    # Test unflattens
    assert np.allclose(unflatten(box2, flattened_box2_sample), box2_sample, atol=1.e-6)
    np.testing.assert_array_equal(
        unflatten(discrete, np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0])), 8
    )
    np.testing.assert_array_equal(
        unflatten(multi_binary, np.array([0, 1, 0, 0, 1, 1, 1])), [0, 1, 0, 0, 1, 1, 1]
    )
    np.testing.assert_array_equal(unflatten(multi_discrete, np.array([0, 3, 1])), [0, 3, 1])
    assert np.allclose(unflatten(d, flattened_d_sample)['first'], d_sample['first'], atol=1.e-8)
    assert np.allclose(unflatten(d, flattened_d_sample)['second'], d_sample['second'], atol=1.e-8)
    assert unflatten(t, flattened_t_sample)[0] == 6
    assert np.allclose(unflatten(t, flattened_t_sample)[1], t_sample[1], atol=1.e-8)
    assert np.allclose(unflatten(t, flattened_t_sample)[2], t_sample[2], atol=1.e-8)
    assert unflatten(combo, flattened_combo_sample)[0]['first'] == 2
    np.testing.assert_array_equal(
        unflatten(combo, flattened_combo_sample)[0]['second'], combo_sample[0]['second']
    )
    np.testing.assert_array_equal(unflatten(combo, flattened_combo_sample)[1], combo_sample[1])


def test_flatten_space():
    flattened_box_space = flatten_space(box)
    assert flattened_box_space == Box(2, 16, (12,), np.int)
    flattened_box2_space = flatten_space(box2)
    assert flattened_box2_space == Box(2.4, 16.1, (12,))
    flattened_discrete_space = flatten_space(discrete)
    assert flattened_discrete_space == Box(0, 1, (11,), np.int)
    flattened_multi_binary_space = flatten_space(multi_binary)
    assert flattened_multi_binary_space == Box(0, 1, (7,), np.int)
    flattened_multi_discrete_space = flatten_space(multi_discrete)
    assert flattened_multi_discrete_space == Box(
        np.array([0, 0, 0]), np.array([2, 6, 4]), (3,), np.int
    )
    flattened_d_space = flatten_space(d)
    assert flattened_d_space == Box(
        np.array([
            2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 0, 0, 0, 0, 0, 0, 0
        ]),
        np.array([
            16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 1, 1, 1, 1, 1,
            1, 1
        ]),
        (19,)
    )
    flattened_t_space = flatten_space(t)
    assert flattened_t_space == Box(
        np.array([
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4, 2.4,
            2.4, 0, 0, 0
        ]),
        np.array([
            1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1, 16.1,
            16.1, 16.1, 16.1, 2, 6, 4
        ]),
        (26,)
    )
    flattened_combo_space = flatten_space(combo)
    assert flattened_combo_space == Box(
        np.array([
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 0, 0, 0, 0, 0, 0,
            0
        ]),
        np.array([
            1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
            16,  1,  1,  1,  1,  1,  1,  1
        ]),
        (30,),
        np.int
    )


# -- Test flatten wrappers --- #
class MultiAgentContinuousGymSpaceSim(MultiAgentSim):
    def __init__(self):
        self.params = {'params': "there are none"}
        self.agents = {
            'agent0': Agent(
                id='agent0',
                observation_space=MultiBinary(4),
                action_space=Tuple((
                    Dict({'first': Discrete(4), 'second': Box(low=-1, high=3, shape=(2,))}),
                    MultiBinary(3)
                ))
            ),
            'agent1': Agent(
                id='agent1',
                observation_space=Box(low=0, high=1, shape=(1,)),
                action_space=MultiDiscrete([4, 6, 2])
            ),
            'agent2': Agent(
                id='agent2',
                observation_space=MultiDiscrete([2, 2]),
                action_space=Dict({'alpha': MultiBinary(3)})
            ),
            'agent3': Agent(
                id='agent3',
                observation_space=Dict({
                    'first': Discrete(4),
                    'second': Box(low=-1, high=3, shape=(2,), dtype=np.int)
                }),
                action_space=Tuple((Discrete(3), MultiDiscrete([10, 10]), Discrete(2)))
            )
        }

    def get_obs(self, agent_id, **kwargs):
        if agent_id == 'agent0':
            return [0, 1, 1, 0]
        elif agent_id == 'agent1':
            return [0.98]
        elif agent_id == 'agent2':
            return [1, 0]
        elif agent_id == 'agent3':
            return {'first': 1, 'second': [-1, 1]}

    def get_info(self, agent_id, **kwargs):
        return self.action[agent_id]


def test_flatten_wrapper():
    sim = MultiAgentContinuousGymSpaceSim()
    wrapped_sim = FlattenWrapper(sim)
    assert wrapped_sim.unwrapped == sim
    for agent_id in wrapped_sim.agents:
        assert isinstance(wrapped_sim.agents[agent_id].observation_space, Box)
        assert isinstance(wrapped_sim.agents[agent_id].action_space, Box)
    sim = wrapped_sim
    sim.reset()

    action_0 = {
        'agent0': ({'first': 2, 'second': [-0.24, 1.9]}, [0, 1, 1]),
        'agent1': [3, 2, 0],
        'agent2': {'alpha': [1, 1, 0]},
        'agent3': (2, np.array([0., 6.]), 1)
    }
    sim.step({
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_0.items()
    })
    np.testing.assert_array_equal(sim.get_obs('agent0'), [0, 1, 1, 0])
    assert sim.get_obs('agent1') == 0.98
    np.testing.assert_array_equal(sim.get_obs('agent2'), [1, 0])
    np.testing.assert_array_equal(sim.get_obs('agent3'), np.array([0, 1, 0, 0, -1, 1]))

    assert sim.get_reward('agent0') == 'Reward from agent0'
    assert sim.get_reward('agent1') == 'Reward from agent1'
    assert sim.get_reward('agent2') == 'Reward from agent2'
    assert sim.get_reward('agent3') == 'Reward from agent3'

    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_done('agent3') == 'Done from agent3'
    assert sim.get_all_done() == "Done from all agents and/or simulation."

    assert sim.get_info('agent0')[0]['first'] == action_0['agent0'][0]['first']
    assert np.allclose(
        sim.get_info('agent0')[0]['second'], action_0['agent0'][0]['second'], atol=1.0e-7
    )
    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_0['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_0['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_0['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_0['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_0['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_0['agent3'][2])


    action_1 = {
        'agent0': ({'first': 0, 'second': [0.6, -0.2]}, [1, 1, 1]),
        'agent1': [1, 5, 1],
        'agent2': {'alpha': [1, 0, 0]},
        'agent3': (1, np.array([9., 4.]), 0)
    }
    sim.step({
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_1.items()
    })
    assert sim.get_info('agent0')[0]['first'] == action_1['agent0'][0]['first']
    assert np.allclose(
        sim.get_info('agent0')[0]['second'], action_1['agent0'][0]['second'], atol=1.0e-7
    )
    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_1['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_1['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_1['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_1['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_1['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_1['agent3'][2])


    action_2 = {
        'agent0': ({'first': 1, 'second': [2.2, 0.98]}, [0, 0, 1]),
        'agent1': [2, 0, 1],
        'agent2': {'alpha': [0, 0, 0]},
        'agent3': (0, np.array([7., 7.]), 0)
    }
    sim.step({
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_2.items()
    })
    assert sim.get_info('agent0')[0]['first'] == action_2['agent0'][0]['first']
    assert np.allclose(
        sim.get_info('agent0')[0]['second'], action_2['agent0'][0]['second'], atol=1.0e-7)

    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_2['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_2['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_2['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_2['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_2['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_2['agent3'][2])



from abmarl.algs.monte_carlo import exploring_starts, epsilon_soft, off_policy
from abmarl.sim.corridor import MultiCorridor as Corridor
from abmarl.managers import AllStepManager
from abmarl.sim.wrappers import RavelDiscreteWrapper
from abmarl.external import GymWrapper
from abmarl.pols import RandomFirstActionPolicy, EpsilonSoftPolicy, GreedyPolicy


def test_exploring_starts_corridor():
    sim = AllStepManager(RavelDiscreteWrapper(Corridor(num_agents=1)))
    sim, q_table, policy = exploring_starts(sim, iteration=100, horizon=10)

    assert isinstance(sim, GymWrapper)
    assert isinstance(sim.sim, AllStepManager)
    assert isinstance(sim.sim.sim, RavelDiscreteWrapper)
    assert isinstance(sim.sim.sim.sim, Corridor)

    assert q_table.shape == (sim.observation_space.n, sim.action_space.n)
    assert isinstance(policy, RandomFirstActionPolicy)


def test_epsilon_soft():
    sim = AllStepManager(RavelDiscreteWrapper(Corridor(num_agents=1)))
    sim, q_table, policy = epsilon_soft(sim, iteration=1000, horizon=20)

    assert isinstance(sim, GymWrapper)
    assert isinstance(sim.sim, AllStepManager)
    assert isinstance(sim.sim.sim, RavelDiscreteWrapper)
    assert isinstance(sim.sim.sim.sim, Corridor)

    assert q_table.shape == (sim.observation_space.n, sim.action_space.n)
    assert isinstance(policy, EpsilonSoftPolicy)

    obs = sim.reset()
    for _ in range(10):
        action = policy.act(obs)
        obs, reward, done, info = sim.step(action)
        if done:
            break
    assert done


def test_off_policy():
    sim = AllStepManager(RavelDiscreteWrapper(Corridor(num_agents=1)))
    sim, q_table, policy = off_policy(sim, iteration=100, horizon=10)

    assert isinstance(sim, GymWrapper)
    assert isinstance(sim.sim, AllStepManager)
    assert isinstance(sim.sim.sim, RavelDiscreteWrapper)
    assert isinstance(sim.sim.sim.sim, Corridor)

    assert q_table.shape == (sim.observation_space.n, sim.action_space.n)
    assert isinstance(policy, GreedyPolicy)

    obs = sim.reset()
    for _ in range(10):
        action = policy.act(obs)
        obs, reward, done, info = sim.step(action)
        if done:
            break
    assert done



# This probably shouldn't go in algs, but we'll move it later after we've figured out the
# architecture a bit more.

def generate_episode(sim, policy, horizon=200):
    """
    Generate an episode from a policy acting on an simulation.

    Returns: sequence of state, action, reward.
    """
    obs = sim.reset()
    policy.reset() # Reset the policy too so that it knows its the beginning of the episode.
    states, actions, rewards = [], [], []
    states.append(obs)
    for _ in range(horizon):
        action = policy.act(obs)
        obs, reward, done, _ = sim.step(action)
        states.append(obs)
        actions.append(action)
        rewards.append(reward)
        if done:
            break

    states.pop() # Pop off the terminating state
    return states, actions, rewards



import numpy as np
import pytest

from abmarl.pols import GreedyPolicy, EpsilonSoftPolicy, RandomFirstActionPolicy


def test_abstract_policy():
    from abmarl.pols.policy import Policy
    with pytest.raises(TypeError):
        Policy(np.zeros((2,3)))


def test_greedy_policy_init():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = GreedyPolicy(table)
    np.testing.assert_array_equal(table, policy.q_table)


def test_greedy_policy_act():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = GreedyPolicy(table)
    action = policy.act(0)
    assert action == np.argmax(policy.q_table[0])


def test_greedy_policy_probability():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = GreedyPolicy(table)
    prob = policy.probability(4, 2)
    assert prob == (1 if 2 == np.argmax(policy.q_table[4]) else 0)


def test_epsilon_soft_policy_init():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = EpsilonSoftPolicy(table)
    np.testing.assert_array_equal(table, policy.q_table)
    assert policy.epsilon == 0.1


def test_epsilon_soft_policy_init_epsilon():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    with pytest.raises(AssertionError):
        EpsilonSoftPolicy(table, epsilon=-0.2)
    with pytest.raises(AssertionError):
        EpsilonSoftPolicy(table, epsilon=1.4)
    EpsilonSoftPolicy(table, epsilon=0.5)


def test_epsilon_soft_policy_act():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,20))
    policy = EpsilonSoftPolicy(table, epsilon=0.)
    action = policy.act(3)
    assert action == np.argmax(policy.q_table[3])


def test_epsilon_soft_policy_probability():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = EpsilonSoftPolicy(table, epsilon=0.5)
    prob = policy.probability(4, 2)
    assert prob == (
        1 - policy.epsilon + policy.epsilon / policy.q_table[4].size
        if 2 == np.argmax(policy.q_table[4])
        else policy.epsilon / policy.q_table[4].size
    )


def test_random_first_action_policy_init():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = RandomFirstActionPolicy(table)
    np.testing.assert_array_equal(table, policy.q_table)


def test_random_first_action_policy_reset():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = RandomFirstActionPolicy(table)
    policy.reset()
    assert policy.take_random_action


def test_random_first_action_policy_no_reset():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = RandomFirstActionPolicy(table)
    with pytest.raises(AttributeError):
        policy.act(5)
    with pytest.raises(AttributeError):
        policy.probability(1, 2)


def test_random_first_action_policy_act():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = RandomFirstActionPolicy(table)

    policy.reset()
    policy.act(1)
    assert not policy.take_random_action
    action = policy.act(2)
    assert action == np.argmax(policy.q_table[2])


def test_random_first_action_policy_probability():
    np.random.seed(24)
    table = np.random.normal(0, 1, size=(6,3))
    policy = RandomFirstActionPolicy(table)

    policy.reset()
    prob = policy.probability(1, 1)
    assert prob == 1. / policy.q_table[1].size

    policy.act(2)
    prob = policy.probability(0, 2)
    assert prob == (1 if 2 == np.argmax(policy.q_table[4]) else 0)



import os
import shutil

main_if_block_begin = """
if __name__ == "__main__":
    # Create output directory and save to params
    import os
    import time
    home = os.path.expanduser("~")
    output_dir = os.path.join(
        home, 'abmarl_results/{}_{}'.format(
            params['experiment']['title'], time.strftime('%Y-%m-%d_%H-%M')
        )
    )
    params['ray_tune']['local_dir'] = output_dir
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Copy this configuration file to the output directory
    import shutil
    shutil.copy(os.path.join(os.getcwd(), __file__), output_dir)

    # Initialize and run ray
    import ray
    from ray import tune
"""

main_if_block_end = """
    tune.run(**params['ray_tune'])
    ray.shutdown()
"""


def _process_magpie_sbatch(config, full_runnable_config_path):
    with open(
        '/usr/tce/packages/magpie/magpie2/submission-scripts/script-sbatch-srun/'
        'magpie.sbatch-srun-ray', 'r'
    ) as file_reader:
        magpie_script = file_reader.readlines()
    for ndx, line in enumerate(magpie_script):
        if line.startswith("#SBATCH --nodes"):
            magpie_script[ndx] = "#SBATCH --nodes={}\n".format(config['nodes'])
        elif line.startswith("#SBATCH --time"):
            magpie_script[ndx] = "#SBATCH --time={}:00:00\n".format(config['time_limit'])
        elif line.startswith("#SBATCH --job-name"):
            magpie_script[ndx] = "#SBATCH --job-name={}\n".format(config['title'])
        elif line.startswith("#SBATCH --partition"):
            magpie_script[ndx] = "#SBATCH --partition=pbatch\n"
        elif line.startswith('export MAGPIE_PYTHON'):
            magpie_script[ndx] = 'export MAGPIE_PYTHON="{}"\n'.format(shutil.which('python'))
        elif line.startswith('export RAY_PATH'):
            magpie_script[ndx] = 'export RAY_PATH="{}"\n'.format(shutil.which('ray'))
        elif line.startswith('export RAY_LOCAL_DIR="/tmp/${USER}/ray"'):
            pass
            # I cannot specify the output directory becasue it doesn't exist
            # until the script is actually run!
        elif line.startswith('export RAY_JOB'):
            magpie_script[ndx] = 'export RAY_JOB="script"\n'
        elif line.startswith('# export RAY_SCRIPT_PATH'):
            magpie_script[ndx] = 'export RAY_SCRIPT_PATH="{}"\n'.format(full_runnable_config_path)
    with open(
        os.path.join(
            os.path.dirname(full_runnable_config_path),
            f"{config['title']}_magpie.sbatch-srun-ray"), 'w'
    ) as file_writer:
        file_writer.writelines(magpie_script)
    return "    ray.init(address=os.environ['MAGPIE_RAY_ADDRESS'])"


def run(full_config_path, parameters):
    """Convert a configuration file to a runnable script, outputting additional
    scripts as requested."""
    # Copy the configuration script
    import os
    import shutil
    full_runnable_config_path = os.path.join(
        os.path.dirname(full_config_path),
        'runnable_' + os.path.basename(full_config_path)
    )
    shutil.copy(full_config_path, full_runnable_config_path)

    ray_init_line = "    ray.init()"
    if parameters.magpie:
        # We need to get two parameters from the experiment configuration. We don't want to load the
        # entire thing because that is overkill and its costly, so we just read the file and store
        # a few pieces.
        with open(full_config_path, 'r') as file_reader:
            config_items_needed = {
                'nodes': parameters.nodes,
                'time_limit': parameters.time_limit,
            }
            for line in file_reader.readlines():
                if line.strip().strip("'").strip('"').startswith('title'):
                    title = line.split(':')[1].strip().strip(',')
                    exec("config_items_needed['title'] = {}".format(title))
                    break
                    # I'm not worried about executing here becuase the entire module will be
                    # executed when the script is run.
        try:
            ray_init_line = _process_magpie_sbatch(config_items_needed, full_runnable_config_path)
        except FileNotFoundError:
            print('Could not find magpie. Is it installed on your HPC system?')

    # Open the runnable file and write parts to enable runnable
    with open(full_runnable_config_path, 'a') as file_writer:
        file_writer.write('\n')
        file_writer.write(main_if_block_begin)
        file_writer.write(ray_init_line)
        file_writer.write(main_if_block_end)



from abmarl.algs.generate_episode import generate_episode


class Sim:
    def reset(self):
        self.count = 0
        return self.count

    def step(self, action):
        self.count = action
        done = True if abs(self.count) > 100 else False
        return self.count, action, done, {}


class Policy:
    def reset(self):
        self.first_guess = True

    def act(self, obs):
        action = obs + 1 if self.first_guess else -2 * obs
        self.first_guess = False
        return action


def test_generate_episode():
    states, actions, rewards = generate_episode(Sim(), Policy())
    assert states == [0, 1, -2, 4, -8, 16, -32, 64]
    assert actions == [1, -2, 4, -8, 16, -32, 64, -128]
    assert rewards == [1, -2, 4, -8, 16, -32, 64, -128]



#!/usr/bin/env python

import argparse
import os

from abmarl.scripts import train_script as train
from abmarl.scripts import analyze_script as analyze
from abmarl.scripts import visualize_script as visualize
from abmarl.scripts import make_runnable_script as runnable

EXAMPLE_USAGE = """
Example usage for training:
    abmarl train my_experiment.py

Example usage for analysis:
    abmarl analyze my_experiment_directory/ my_analysis_script.py

Example usage for visualizing:
    abmarl visualize my_experiment_directory/ --some-args

Example usage for converting to runnable script:
    abmarl make-runnable my_experiment.py --some-args
"""


def cli():
    parser = argparse.ArgumentParser(
        prog='abmarl',
        description="Train, analyze, and visualize MARL policies.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=EXAMPLE_USAGE
    )
    subparsers = parser.add_subparsers(dest='command')

    train.create_parser(subparsers)
    analyze.create_parser(subparsers)
    visualize.create_parser(subparsers)
    runnable.create_parser(subparsers)
    parameters = parser.parse_args()
    path_config = os.path.join(os.getcwd(), parameters.configuration)

    if parameters.command == 'train':
        train.run(path_config)
    elif parameters.command == 'analyze':
        full_path_subscript = os.path.join(os.getcwd(), parameters.subscript)
        analyze.run(path_config, full_path_subscript, parameters)
    elif parameters.command == 'visualize':
        visualize.run(path_config, parameters)
    elif parameters.command == 'make-runnable':
        runnable.run(path_config, parameters)
    else:
        parser.print_help()



from enum import IntEnum

from gym.spaces import Box, Discrete, MultiBinary
import numpy as np

from abmarl.sim import Agent, AgentBasedSimulation


class MultiCorridor(AgentBasedSimulation):
    """
    MultiCorridor simulation used for testing. Multiple agents spawn along
    a corridor and can choose to move left, right, or stay still. The agents
    must learn to move to the right until they reach the end position. The agent
    cannot move to spaces that area already occupied by other agents. An agent
    is done once it moves to the end position.

    The agent can observe its own position. It can also see if the two squares
    near it are occupied.
    """
    class Actions(IntEnum):
        LEFT = 0
        STAY = 1
        RIGHT = 2

    def __init__(self, end=10, num_agents=5):
        self.end = end
        agents = {}
        for i in range(num_agents):
            agents[f'agent{i}'] = Agent(
                id=f'agent{i}',
                action_space=Discrete(3),
                observation_space={
                    'position': Box(0, self.end-1, (1,), np.int),
                    'left': MultiBinary(1),
                    'right': MultiBinary(1)
                }
            )
        self.agents = agents

        self.finalize()

    def reset(self, **kwargs):
        """
        Randomly locate the agents on unique spaces within the Corridor.
        """
        location_sample = np.random.choice(self.end-1, len(self.agents), False)
        self.corridor = np.empty(self.end, dtype=object)
        for i, agent in enumerate(self.agents.values()):
            agent.position = location_sample[i]
            self.corridor[location_sample[i]] = agent

        # Track the agents' rewards over multiple steps.
        self.reward = {agent_id: 0 for agent_id in self.agents}

    def step(self, action_dict, **kwargs):
        """
        The agents can choose to move left, move right, or stay where they are. If
        an agent bumps into another agent, then both agents receive a penalty.
        The offending agent receives a larger penalty than the offended agent.
        The agent is done when it reaches the end of the corridor.
        """
        for agent_id, action in action_dict.items():
            agent = self.agents[agent_id]
            if action == self.Actions.LEFT:
                if agent.position != 0 and self.corridor[agent.position-1] is None:
                    # Good move, no extra penalty
                    self.corridor[agent.position] = None
                    agent.position -= 1
                    self.corridor[agent.position] = agent
                    self.reward[agent_id] -= 1 # Entropy penalty
                elif agent.position == 0: # Tried to move left from left-most square
                    # Bad move, only acting agent is involved and should be penalized.
                    self.reward[agent_id] -= 5 # Bad move
                else: # There was another agent to the left of me that I bumped into
                    # Bad move involving two agents. Both are penalized
                    self.reward[agent_id] -= 5 # Penalty for offending agent
                    self.reward[self.corridor[agent.position-1].id] -= 2 # Penalty for offended
            elif action == self.Actions.RIGHT:
                if self.corridor[agent.position + 1] is None:
                    # Good move, but is the agent done?
                    self.corridor[agent.position] = None
                    agent.position += 1
                    if agent.position == self.end-1:
                        # Agent has reached the end of the corridor!
                        self.reward[agent_id] += self.end ** 2
                    else:
                        # Good move, no extra penalty
                        self.corridor[agent.position] = agent
                        self.reward[agent_id] -= 1 # Entropy penalty
                else: # There was another agent to the right of me that I bumped into
                    # Bad move involving two agents. Both are penalized
                    self.reward[agent_id] -= 5 # Penalty for offending agent
                    self.reward[self.corridor[agent.position+1].id] -= 2 # Penalty for offended
            elif action == self.Actions.STAY:
                self.reward[agent_id] -= 1 # Entropy penalty

    def render(self, *args, fig=None, **kwargs):
        """
        Visualize the state of the simulation. If a figure is received, then we
        will draw but not actually plot because we assume the caller will do the
        work (e.g. with an Animation object). If there is no figure received, then
        we will draw and plot the simulation.
        """
        draw_now = fig is None
        if draw_now:
            from matplotlib import pyplot as plt
            fig = plt.gcf()

        fig.clear()
        ax = fig.gca()
        ax.set(xlim=(-0.5, self.end + 0.5), ylim=(-0.5, 0.5))
        ax.set_xticks(np.arange(-0.5, self.end + 0.5, 1.))
        ax.scatter(np.array(
            [agent.position for agent in self.agents.values()]),
            np.zeros(len(self.agents)),
            marker='s', s=200, c='g'
        )

        if draw_now:
            plt.plot()
            plt.pause(1e-17)

    def get_obs(self, agent_id, **kwargs):
        """
        Agents observe their own position and if the squares to the left and right
        are occupied by other agents.
        """
        agent_position = self.agents[agent_id].position
        if agent_position == 0 or self.corridor[agent_position-1] is None:
            left = False
        else:
            left = True
        if agent_position == self.end-1 or self.corridor[agent_position+1] is None:
            right = False
        else:
            right = True
        return {
            'position': [agent_position],
            'left': [left],
            'right': [right],
        }

    def get_done(self, agent_id, **kwargs):
        """
        Agents are done when they reach the end of the corridor.
        """
        return self.agents[agent_id].position == self.end - 1

    def get_all_done(self, **kwargs):
        """
        Simulation is done when all agents have reached the end of the corridor.
        """
        for agent in self.agents.values():
            if agent.position != self.end - 1:
                return False
        return True

    def get_reward(self, agent_id, **kwargs):
        """
        The agent's reward is tracked throughout the simulation and returned here.
        """
        agent_reward = self.reward[agent_id]
        self.reward[agent_id] = 0
        return agent_reward

    def get_info(self, agent_id, **kwargs):
        """
        Just return an empty dictionary becuase this simulation does not track
        any info.
        """
        return {}



from ray.rllib import MultiAgentEnv


class MultiAgentWrapper(MultiAgentEnv):
    """
    Enable connection between SimulationManager and RLlib Trainer.

    Wraps a SimulationManager and forwards all calls to the manager. This class
    is boilerplate and needed because RLlib checks that the simulation is an instance
    of MultiAgentEnv.

    Attributes:
        sim: The SimulationManager.
    """
    def __init__(self, sim):
        from abmarl.managers import SimulationManager
        assert isinstance(sim, SimulationManager)
        self.sim = sim

    def reset(self):
        """See SimulationManager."""
        return self.sim.reset()

    def step(self, actions):
        """See SimulationManager."""
        return self.sim.step(actions)

    def render(self, *args, **kwargs):
        """See SimulationManager."""
        return self.sim.render(*args, **kwargs)

    @property
    def unwrapped(self):
        """
        Fall through all the wrappers to the SimulationManager.

        Returns:
            The wrapped SimulationManager.
        """
        try:
            return self.sim.unwrapped
        except AttributeError:
            return self.sim



from abmarl.sim import AgentBasedSimulation


class Wrapper(AgentBasedSimulation):
    """
    Abstract Wrapper class implements the AgentBasedSimulation interface. The simulation
    is stored and the simulation agents are deep-copied. The interface functions
    calls are forwarded to the simulation.
    """
    def __init__(self, sim):
        """
        Wrap the simulation and copy the agents.
        """
        assert isinstance(sim, AgentBasedSimulation)
        self.sim = sim

        import copy
        self.agents = copy.deepcopy(sim.agents)

    def reset(self, **kwargs):
        self.sim.reset(**kwargs)

    def step(self, action, **kwargs):
        self.sim.step(action, **kwargs)

    def render(self, **kwargs):
        self.sim.render(**kwargs)

    def get_obs(self, agent_id, **kwargs):
        return self.sim.get_obs(agent_id, **kwargs)

    def get_reward(self, agent_id, **kwargs):
        return self.sim.get_reward(agent_id, **kwargs)

    def get_done(self, agent_id, **kwargs):
        return self.sim.get_done(agent_id, **kwargs)

    def get_all_done(self, **kwargs):
        return self.sim.get_all_done(**kwargs)

    def get_info(self, agent_id, **kwargs):
        return self.sim.get_info(agent_id, **kwargs)

    @property
    def unwrapped(self):
        """
        Fall through all the wrappers and obtain the original, completely unwrapped simulation.
        """
        try:
            return self.sim.unwrapped
        except AttributeError:
            return self.sim



from gym import Env as GymEnv


class GymWrapper(GymEnv):
    """
    Wrap an AgentBasedSimulation object with only a single agent to the gym.Env interface.
    This wrapper exposes the single agent's observation and action space directly
    in the simulation.
    """
    def __init__(self, sim):
        from abmarl.managers import SimulationManager
        assert isinstance(sim, SimulationManager)
        assert len(sim.agents) == 1 # Can only work with single agents
        self.sim = sim
        self.agent_id, agent = next(iter(sim.agents.items()))
        self.observation_space = agent.observation_space
        self.action_space = agent.action_space

    def reset(self, **kwargs):
        """
        Return the observation from the single agent.
        """
        obs = self.sim.reset(**kwargs)
        return obs[self.agent_id]

    def step(self, action, **kwargs):
        """
        Wrap the action by storing it in a dict that maps the agent's id to the
        action. Pass to sim.step. Return the observation, reward, done, and
        info from the single agent.
        """
        obs, reward, done, info = self.sim.step({self.agent_id: action}, **kwargs)
        return obs[self.agent_id], \
            reward[self.agent_id], \
            done[self.agent_id], \
            info[self.agent_id]

    def render(self, **kwargs):
        """
        Forward render calls to the composed simulation.
        """
        self.sim.render(**kwargs)

    @property
    def unwrapped(self):
        """
        Fall through all the wrappers and obtain the original, completely unwrapped simulation.
        """
        try:
            return self.sim.unwrapped
        except AttributeError:
            return self.sim



from abmarl.sim import AgentBasedSimulation
from abmarl.sim import Agent

from gym.spaces import Discrete, MultiBinary, MultiDiscrete, Box, Dict, Tuple
import numpy as np


class FillInHelper(AgentBasedSimulation):
    def reset(self):
        pass

    def step(self, action):
        pass

    def render(self):
        pass

    def get_obs(self, agent_id, **kwargs):
        pass

    def get_reward(self, agent_id, **kwargs):
        pass

    def get_done(self, agent_id, **kwargs):
        pass

    def get_all_done(self, **kwargs):
        pass

    def get_info(self, agent_id, **kwargs):
        pass


class MultiAgentSim(FillInHelper):
    def __init__(self, num_agents=3):
        self.agents = {
            'agent' + str(i): Agent(
                id='agent'+str(i), observation_space=Discrete(2), action_space=Discrete(2)
            ) for i in range(num_agents)
        }

    def reset(self):
        self.action = {agent.id: None for agent in self.agents.values()}

    def step(self, action_dict):
        for agent_id, action in action_dict.items():
            self.action[agent_id] = action

    def get_obs(self, agent_id, **kwargs):
        return "Obs from " + agent_id

    def get_reward(self, agent_id, **kwargs):
        return "Reward from " + agent_id

    def get_done(self, agent_id, **kwargs):
        return "Done from " + agent_id

    def get_all_done(self, **kwargs):
        return "Done from all agents and/or simulation."

    def get_info(self, agent_id, **kwargs):
        return {'Action from ' + agent_id: self.action[agent_id]}


class MultiAgentGymSpacesSim(MultiAgentSim):
    def __init__(self):
        self.params = {'params': "there are none"}
        self.agents = {
            'agent0': Agent(
                id='agent0',
                observation_space=MultiBinary(4),
                action_space=Tuple((
                    Dict({
                        'first': Discrete(4),
                        'second': Box(low=-1, high=3, shape=(2,), dtype=np.int)
                    }),
                    MultiBinary(3)
                ))
            ),
            'agent1': Agent(
                id='agent1',
                observation_space=Box(low=0, high=1, shape=(1,), dtype=np.int),
                action_space=MultiDiscrete([4, 6, 2])
            ),
            'agent2': Agent(
                id='agent2',
                observation_space=MultiDiscrete([2, 2]),
                action_space=Dict({'alpha': MultiBinary(3)})
            ),
            'agent3': Agent(
                id='agent3',
                observation_space=Dict({
                    'first': Discrete(4),
                    'second': Box(low=-1, high=3, shape=(2,), dtype=np.int)
                }),
                action_space=Tuple((Discrete(3), MultiDiscrete([10, 10]), Discrete(2)))
            )
        }


    def get_obs(self, agent_id, **kwargs):
        if agent_id == 'agent0':
            return [0, 0, 0, 1]
        elif agent_id == 'agent1':
            return 0
        elif agent_id == 'agent2':
            return [1, 0]
        elif agent_id == 'agent3':
            return {'first': 1, 'second': [3, 1]}


    def get_info(self, agent_id, **kwargs):
        return self.action[agent_id]



from abmarl.sim.wrappers.ravel_discrete_wrapper import ravel, unravel
from abmarl.sim.wrappers import RavelDiscreteWrapper
from abmarl.sim import Agent

from gym.spaces import MultiDiscrete, Discrete, MultiBinary, Box, Dict, Tuple
import numpy as np
import pytest
from .helpers import FillInHelper, MultiAgentGymSpacesSim


def test_ravel():
    my_space = Dict({
        'a': MultiDiscrete([5, 3]),
        'b': MultiBinary(4),
        'c': Box(np.array([[-2, 6, 3],[0, 0, 1]]), np.array([[2, 12, 5],[2, 4, 2]]), dtype=np.int),
        'd': Dict({
            1: Discrete(3),
            2: Box(1, 3, (2,), np.int)
        }),
        'e': Tuple((
            MultiDiscrete([4, 1, 5]),
            MultiBinary(2),
            Dict({
                'my_dict': Discrete(11)
            })
        )),
        'f': Discrete(6),
    })
    point = {
        'a': [3, 1],
        'b': [0, 1, 1, 0],
        'c': np.array([[0, 7, 5],[1, 3, 1]]),
        'd': {1: 2, 2: np.array([1, 3])},
        'e': ([1,0,4], [1, 1], {'my_dict': 5}),
        'f': 1
    }
    ravelled_point = ravel(my_space, point)
    unravelled_point = unravel(my_space, ravelled_point)

    assert ravelled_point == 74748022765
    np.testing.assert_array_equal(unravelled_point['a'], point['a'])
    np.testing.assert_array_equal(unravelled_point['b'], point['b'])
    np.testing.assert_array_equal(unravelled_point['c'], point['c'])
    assert unravelled_point['d'][1] == point['d'][1]
    np.testing.assert_array_equal(unravelled_point['d'][2], point['d'][2])
    np.testing.assert_array_equal(unravelled_point['e'][0], point['e'][0])
    np.testing.assert_array_equal(unravelled_point['e'][1], point['e'][1])
    np.testing.assert_array_equal(unravelled_point['e'][2], point['e'][2])
    assert unravelled_point['f'] == point['f']


# Observations that we don't support
class FloatObservation(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0', observation_space=Box(-1.0, 1.0, shape=(4,)), action_space=Discrete(3)
        )}


class UnboundedBelowObservation(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0', observation_space=Box(
                np.array([0, 13, -3, -np.inf]),
                np.array([0, 20, 0, 0]),
                dtype=np.int
            ),
            action_space=Discrete(3)
        )}


class UnboundedAboveObservation(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0', observation_space=Box(
                np.array([0, 12, 20, 0]),
                np.array([np.inf, 20, 24, np.inf]),
                dtype=np.int
            ),
            action_space=Discrete(2)
        )}


# Actions that we don't support
class FloatAction(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0', observation_space=Box(-1.0, 1.0, shape=(4,)), action_space=Discrete(3)
        )}


class UnboundedBelowAction(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0',
            observation_space=Box(
                np.array([0, 13, -3, -np.inf]),
                np.array([0, 20, 0, 0]),
                dtype=np.int
            ),
            action_space=Discrete(3)
        )}


class UnboundedAboveAction(FillInHelper):
    def __init__(self):
        self.agents = {'agent0': Agent(
            id='agent0',
            observation_space=Box(
                np.array([0, 12, 20, 0]),
                np.array([np.inf, 20, 24, np.inf]),
                dtype=np.int
            ),
            action_space=Discrete(2)
        )}


def test_exceptions():
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(FloatObservation())
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(UnboundedBelowObservation())
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(UnboundedAboveObservation())
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(FloatAction())
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(UnboundedBelowAction())
    with pytest.raises(AssertionError):
        RavelDiscreteWrapper(UnboundedAboveAction())


def test_ravel_wrapper():
    sim = MultiAgentGymSpacesSim()
    wrapped_sim = RavelDiscreteWrapper(sim)
    assert wrapped_sim.unwrapped == sim
    for agent_id in wrapped_sim.agents:
        assert isinstance(wrapped_sim.agents[agent_id].observation_space, Discrete)
        assert isinstance(wrapped_sim.agents[agent_id].action_space, Discrete)
    sim = wrapped_sim

    sim.reset()
    assert sim.get_obs('agent0') == 1
    assert sim.get_obs('agent1') == 0
    assert sim.get_obs('agent2') == 2
    assert sim.get_obs('agent3') == 47

    action_0 = {
        'agent0': ({'first': 2, 'second': [-1, 2]}, [0, 1, 1]),
        'agent1': [3, 2, 0],
        'agent2': {'alpha': [1, 1, 0]},
        'agent3': (2, np.array([0, 6]), 1)
    }
    action_0_wrapped = {
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_0.items()
    }

    action_1 = {
        'agent0': ({'first': 0, 'second': [3, 3]}, [1, 1, 1]),
        'agent1': [1, 5, 1],
        'agent2': {'alpha': [1, 0, 0]},
        'agent3': (1, np.array([9, 4]), 0)
    }
    action_1_wrapped = {
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_1.items()
    }

    action_2 = {
        'agent0': ({'first': 1, 'second': [1, 0]}, [0, 0, 1]),
        'agent1': [2, 0, 1],
        'agent2': {'alpha': [0, 0, 0]},
        'agent3': (0, np.array([7, 7]), 0)
    }
    action_2_wrapped = {
        agent_id: sim.unwrap_action(sim.sim.agents[agent_id], action)
        for agent_id, action in action_2.items()
    }

    sim.step(action_0_wrapped)
    assert sim.get_obs('agent0') == 1
    assert sim.get_obs('agent1') == 0
    assert sim.get_obs('agent2') == 2
    assert sim.get_obs('agent3') == 47

    assert sim.get_reward('agent0') == 'Reward from agent0'
    assert sim.get_reward('agent1') == 'Reward from agent1'
    assert sim.get_reward('agent2') == 'Reward from agent2'
    assert sim.get_reward('agent3') == 'Reward from agent3'

    assert sim.get_done('agent0') == 'Done from agent0'
    assert sim.get_done('agent1') == 'Done from agent1'
    assert sim.get_done('agent2') == 'Done from agent2'
    assert sim.get_done('agent3') == 'Done from agent3'

    assert sim.get_info('agent0')[0]['first'] == action_0['agent0'][0]['first']
    np.testing.assert_array_equal(
        sim.get_info('agent0')[0]['second'], action_0['agent0'][0]['second']
    )
    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_0['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_0['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_0['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_0['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_0['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_0['agent3'][2])

    sim.step(action_1_wrapped)
    assert sim.get_info('agent0')[0]['first'] == action_1['agent0'][0]['first']
    np.testing.assert_array_equal(
        sim.get_info('agent0')[0]['second'], action_1['agent0'][0]['second']
    )
    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_1['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_1['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_1['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_1['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_1['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_1['agent3'][2])

    sim.step(action_2_wrapped)
    assert sim.get_info('agent0')[0]['first'] == action_2['agent0'][0]['first']
    np.testing.assert_array_equal(
        sim.get_info('agent0')[0]['second'], action_2['agent0'][0]['second']
    )
    np.testing.assert_array_equal(sim.get_info('agent0')[1], action_2['agent0'][1])
    np.testing.assert_array_equal(sim.get_info('agent1'), action_2['agent1'])
    np.testing.assert_array_equal(sim.get_info('agent2')['alpha'], action_2['agent2']['alpha'])
    np.testing.assert_array_equal(sim.get_info('agent3')[0], action_2['agent3'][0])
    np.testing.assert_array_equal(sim.get_info('agent3')[1], action_2['agent3'][1])
    np.testing.assert_array_equal(sim.get_info('agent3')[2], action_2['agent3'][2])




from matplotlib import pyplot as plt
import numpy as np

from abmarl.sim.gridworld.base import GridWorldSimulation
from abmarl.sim.gridworld.agent import GridObservingAgent, MovingAgent, GridWorldAgent
from abmarl.sim.gridworld.state import PositionState
from abmarl.sim.gridworld.actor import MoveActor
from abmarl.sim.gridworld.observer import SingleGridObserver


class MazeNavigationAgent(GridObservingAgent, MovingAgent):
    def __init__(self, **kwargs):
        super().__init__(move_range=1, **kwargs)


class MazeNaviationSim(GridWorldSimulation):
    def __init__(self, **kwargs):
        self.agents = kwargs['agents']
        self.navigator = kwargs['agents']['navigator']
        self.target = kwargs['agents']['target']

        # State Components
        self.position_state = PositionState(**kwargs)

        # Action Components
        self.move_actor = MoveActor(**kwargs)

        # Observation Components
        self.grid_observer = SingleGridObserver(**kwargs)

        self.finalize()

    def reset(self, **kwargs):
        self.position_state.reset(**kwargs)

        # Track the rewards
        self.reward = 0

    def step(self, action_dict, **kwargs):
        # Process moves
        action = action_dict['navigator']
        move_result = self.move_actor.process_action(self.navigator, action, **kwargs)
        if not move_result:
            self.reward -= 0.1

        # Entropy penalty
        self.reward -= 0.01

    def get_obs(self, agent_id, **kwargs):
        return {
            **self.grid_observer.get_obs(self.navigator, **kwargs)
        }

    def get_reward(self, agent_id, **kwargs):
        if self.get_all_done():
            self.reward = 1
        reward = self.reward
        self.reward = 0
        return reward

    def get_done(self, agent_id, **kwargs):
        return self.get_all_done()

    def get_all_done(self, **kwargs):
        return np.all(self.navigator.position == self.target.position)

    def get_info(self, agent_id, **kwargs):
        return {}


if __name__ == "__main__":
    object_registry = {
        'N': lambda n: MazeNavigationAgent(
            id='navigator',
            encoding=1,
            view_range=2,
            render_color='blue',
        ),
        'T': lambda n: GridWorldAgent(
            id='target',
            encoding=3,
            render_color='green'
        ),
        'W': lambda n: GridWorldAgent(
            id=f'wall{n}',
            encoding=2,
            blocking=True,
            render_shape='s'
        )
    }

    file_name = 'maze.txt'
    sim = MazeNaviationSim.build_sim_from_file(
        file_name,
        object_registry,
        overlapping={1: [3], 3: [1]}
    )
    sim.reset()
    fig = plt.figure()
    sim.render(fig=fig)

    for i in range(100):
        action = {'navigator': sim.navigator.action_space.sample()}
        sim.step(action)
        sim.render(fig=fig)
        done = sim.get_all_done()
        if done:
            plt.pause(1)
            break




from abc import ABC, abstractmethod

import numpy as np

from abmarl.sim.gridworld.base import GridWorldBaseComponent
from abmarl.sim.gridworld.agent import HealthAgent


class StateBaseComponent(GridWorldBaseComponent, ABC):
    """
    Abstract State Component base from which all state components will inherit.
    """
    @abstractmethod
    def reset(self, **kwargs):
        """
        Resets the part of the state for which it is responsible.
        """
        pass


class PositionState(StateBaseComponent):
    """
    Manage the agents' positions in the grid.
    """
    def reset(self, **kwargs):
        """
        Give agents their starting positions.

        We use the agent's initial position if it exists. Otherwise, we randomly
        place the agents in the grid.
        """
        self.grid.reset()
        # Prioritize placing agents with initial positions. We must keep track
        # of which positions have been taken so that
        # the random placement below doesn't try to place an agent there.
        ravelled_positions_available = set(i for i in range(self.rows * self.cols))
        for agent in self.agents.values():
            if agent.initial_position is not None:
                r, c = agent.initial_position
                assert self.grid.place(agent, (r, c)), "All initial positions must " + \
                    "be unique or agents with the same initial positions must be overlappable."
                try:
                    ravelled_positions_available.remove(
                        np.ravel_multi_index(agent.position, (self.rows, self.cols))
                    )
                except KeyError:
                    continue

        # Now place all the rest of the agents who did not have initial positions
        # and block off those positions as well. We have to do this one agent at
        # a time because the list of available positions is updated after each
        # agent is placed.
        for agent in self.agents.values():
            if agent.initial_position is None:
                n = np.random.choice([*ravelled_positions_available], 1)
                r, c = np.unravel_index(n.item(), shape=(self.rows, self.cols))
                assert self.grid.place(agent, (r, c))
                ravelled_positions_available.remove(n.item())


class HealthState(StateBaseComponent):
    """
    Manage the state of the agents' healths.

    Every HealthAgent has a health. If that health falls to zero, that agent dies
    and is remove from the grid.
    """
    def reset(self, **kwargs):
        """
        Give HealthAgents their starting healths.

        We use the agent's initial health if it exists. Otherwise, we randomly
        assign a value between 0 and 1.
        """
        for agent in self.agents.values():
            if isinstance(agent, HealthAgent):
                if agent.initial_health is not None:
                    agent.health = agent.initial_health
                else:
                    agent.health = np.random.uniform(0, 1)




from abc import ABC, abstractmethod

from abmarl.sim.gridworld.base import GridWorldBaseComponent


class DoneBaseComponent(GridWorldBaseComponent, ABC):
    """
    Abstract Done Component class from which all Done Components will inherit.
    """
    @abstractmethod
    def get_done(self, agent, **kwargs):
        """
        Determine if an agent is done in this step.

        Args:
            agent: The agent we are querying.

        Returns:
            True if the agent is done, otherwise False.
        """
        pass

    @abstractmethod
    def get_all_done(self, **kwargs):
        """
        Determine if all the agents are done and/or if the simulation is done.

        Returns:
            True if all agents are done or if the simulation is done. Otherwise
            False.
        """
        pass


class ActiveDone(DoneBaseComponent):
    """
    Inactive agents are indicated as done.
    """
    def get_done(self, agent, **kwargs):
        """
        Return True if the agent is inactive. Otherwise, return False.
        """
        return not agent.active

    def get_all_done(self, **kwargs):
        """
        Return True if all agents are inactive. Otherwise, return False.
        """
        for agent in self.agents.values():
            if agent.active:
                return False
        return True


class OneTeamRemainingDone(ActiveDone):
    """
    Inactive agents are indicated as done.

    If the only active agents are those who are all of the same encoding, then
    the simulation ends.
    """
    def get_all_done(self, **kwargs):
        """
        Return true if all active agents have the same encoding. Otherwise,
        return false.
        """
        encodings = set(agent.encoding for agent in self.agents.values() if agent.active)
        return len(encodings) <= 1
